import io
import random
import subprocess
import sys
import unittest

import pyrtl
from pyrtl.corecircuits import _basic_add
from pyrtl.importexport import _VerilogSanitizer
from pyrtl.rtllib import testingutils as utils

full_adder_blif = r"""
# Generated by Yosys 0.3.0+ (git sha1 7e758d5, clang 3.4-1ubuntu3 -fPIC -Os)
.model full_adder
.inputs x y cin
.outputs sum cout
.names $false
.names $true
1
.names y $not$FA.v:12$3_Y
0 1
.names x $not$FA.v:11$1_Y
0 1
.names cin $not$FA.v:15$6_Y
0 1
.names ind3 ind4 sum
1- 1
-1 1
.names $not$FA.v:15$6_Y ind2 ind3
11 1
.names x $not$FA.v:12$3_Y ind1
11 1
.names ind2 $not$FA.v:16$8_Y
0 1
.names cin $not$FA.v:16$8_Y ind4
11 1
.names x y $and$FA.v:19$11_Y
11 1
.names ind0 ind1 ind2
1- 1
-1 1
.names cin ind2 $and$FA.v:19$12_Y
11 1
.names $and$FA.v:19$11_Y $and$FA.v:19$12_Y cout
1- 1
-1 1
.names $not$FA.v:11$1_Y y ind0
11 1
.end
"""

state_machine_blif = r"""
# Generated by Yosys 0.5+     420 (git sha1 1d62f87, clang 7.0.2 -fPIC -Os)

.model statem
.inputs clk in reset
.outputs out[0] out[1] out[2] out[3]
.names $false
.names $true
1
.names $undef
.names in state[2] $abc$129$n11_1
11 1
.names $abc$129$n11_1 state[3] $auto$fsm_map.cc:238:map_fsm$30[0]
1- 1
-1 1
.names state[2] $abc$129$n13
0 1
.names state[0] $abc$129$n14_1
0 1
.names state[2] state[1] $abc$129$n15
00 1
.names $abc$129$n15 $abc$129$n14_1 $abc$129$n13 out[0]
-00 1
0-0 1
.names state[1] $abc$129$n17
0 1
.names $abc$129$n15 $abc$129$n14_1 $abc$129$n17 out[1]
-00 1
0-0 1
.names $abc$129$n15 $abc$129$n14_1 out[2]
11 1
.names in $abc$129$n13 $auto$fsm_map.cc:118:implement_pattern_cache$38
00 1
# .subckt $_DFF_PP1_ C=clk D=$auto$fsm_map.cc:238:map_fsm$30[0] Q=state[0] R=reset
# .subckt $_DFF_PP0_ C=clk D=$auto$fsm_map.cc:118:implement_pattern_cache$38 Q=state[1] R=reset
# .subckt $_DFF_PP0_ C=clk D=state[0] Q=state[2] R=reset
# .subckt $_DFF_PP0_ C=clk D=state[1] Q=state[3] R=reset
.names $false out[3]
1 1
.end
"""  # noqa: E501

# Manually set the .latch's init values from 2 to arbitrary non-1 numbers, for testing.
# Should result in the same logic, but allows for testing the parser.
counter4bit_blif = r"""
# Generated by Yosys 0.9 (git sha1 UNKNOWN, clang 11.0.0 -fPIC -Os)

.model counter
.inputs clk rst en
.outputs count[0] count[1] count[2] count[3]
.names $false
.names $true
1
.names $undef
.names count[0] $add$counter.v:10$2_Y[0] en $procmux$3_Y[0]
1-0 1
-11 1
.names count[1] $add$counter.v:10$2_Y[1] en $procmux$3_Y[1]
1-0 1
-11 1
.names count[2] $add$counter.v:10$2_Y[2] en $procmux$3_Y[2]
1-0 1
-11 1
.names count[3] $add$counter.v:10$2_Y[3] en $procmux$3_Y[3]
1-0 1
-11 1
.names $procmux$3_Y[0] $false rst $0\count[3:0][0]
1-0 1
-11 1
.names $procmux$3_Y[1] $false rst $0\count[3:0][1]
1-0 1
-11 1
.names $procmux$3_Y[2] $false rst $0\count[3:0][2]
1-0 1
-11 1
.names $procmux$3_Y[3] $false rst $0\count[3:0][3]
1-0 1
-11 1
.latch $0\count[3:0][0] count[0] re clk 2
.latch $0\count[3:0][1] count[1] re clk 0
.latch $0\count[3:0][2] count[2] re clk 3
.latch $0\count[3:0][3] count[3] re clk
.names count[1] count[0] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[1]
11 1
.names count[2] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[1] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[2]
11 1
.names count[1] count[0] $add$counter.v:10$2_Y[1]
10 1
01 1
.names count[2] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[1] $add$counter.v:10$2_Y[2]
10 1
01 1
.names count[3] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[2] $add$counter.v:10$2_Y[3]
10 1
01 1
.names count[0] $true $add$counter.v:10$2_Y[0]
10 1
01 1
.names count[0] $techmap$add$counter.v:10$2.$auto$alumacc.cc:474:replace_alu$53.lcu.g[0]
1 1
.end
"""  # noqa

blif_with_output_as_arg = r"""
# Generated by Yosys 0.9+2406 (git sha1 aee43936, clang 11.0.3 -fPIC -Os)

.model Top
.inputs clk in[0] in[1]
.outputs out
.names $false
.names $true
1
.names $undef
.names out $techmap$add$test.v:9$6.$auto$alumacc.cc:485:replace_alu$60.X[0]
0 1
.names in[0] $not$test.v:6$1_Y[0]
0 1
.names r[0] $not$test.v:6$3_Y[0]
0 1
.latch $techmap$add$test.v:9$6.$auto$alumacc.cc:485:replace_alu$60.X[0] r[0] re clk 2
.names $not$test.v:6$1_Y[0] $not$test.v:6$3_Y[0] out
10 1
01 1
.names $true $not$test.v:6$3_Y[1]
1 1
.names $false $techmap$add$test.v:9$6.$auto$alumacc.cc:485:replace_alu$60.X[1]
1 1
.names $false $techmap$add$test.v:9$6.$auto$alumacc.cc:485:replace_alu$60.X[2]
1 1
.end
"""

simple_unmerged_io_blif = r"""
# Generated by Yosys 0.9+2406 (git sha1 aee43936, clang 11.0.3 -fPIC -Os)

.model top
.inputs a[0] a[1] a[2] a[3]
.outputs b[0] b[1]
.names $false
.names $true
1
.names $undef
.names a[0] b[0]
1 1
.names a[2] b[1]
1 1
.end
"""

four_bit_adder_multi_module = r"""
# Generated by Yosys 0.9+2406 (git sha1 aee43936, clang 11.0.3 -fPIC -Os)

.model four_bit_adder
.inputs a[0] a[1] a[2] a[3] b[0] b[1] b[2] b[3] cin
.outputs s[0] s[1] s[2] s[3] cout
.names $false
.names $true
1
.names $undef
.subckt full_adder a=a[0] b=b[0] cin=cin cout=cout0 s=s[0]
.subckt full_adder a=a[1] b=b[1] cin=cout0 cout=cout1 s=s[1]
.subckt full_adder a=a[2] b=b[2] cin=cout1 cout=cout2 s=s[2]
.subckt full_adder a=a[3] b=b[3] cin=cout2 cout=cout s=s[3]
.end

.model full_adder
.inputs a b cin
.outputs s cout
.names $false
.names $true
1
.names $undef
.names a b $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.X[0]
10 1
01 1
.names a b $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.CO[0]
11 1
.names $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.CO[0] $techmap$add$four_bit_adder_subsub.v:34$2.$auto$alumacc.cc:485:replace_alu$49.CO[0] x
10 1
01 1
.names cin $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.X[0] s
10 1
01 1
.names cin $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.X[0] $techmap$add$four_bit_adder_subsub.v:34$2.$auto$alumacc.cc:485:replace_alu$49.CO[0]
11 1
.subckt pass_through x=x y=y
.subckt pass_through x=y y=cout
.names $false $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.CO[1]
1 1
.names $false $techmap$add$four_bit_adder_subsub.v:34$1.$auto$alumacc.cc:485:replace_alu$46.X[1]
1 1
.end

.model pass_through
.inputs x
.outputs y
.names $false
.names $true
1
.names $undef
.names x y
1 1
.end
"""  # noqa

clock_passing_blif = r"""
# Generated by Yosys 0.9+2406 (git sha1 aee43936, clang 11.0.3 -fPIC -Os)

.model top
.inputs clk a[0] a[1] a[2] a[3] b[0] b[1] b[2] b[3]
.outputs c[0] c[1] c[2] c[3]
.names $false
.names $true
1
.names $undef
.names b[0] w2[0] c[0]
11 1
.names b[1] w2[1] c[1]
11 1
.names b[2] w2[2] c[2]
11 1
.names b[3] w2[3] c[3]
11 1
.subckt passthrough a[0]=a[0] a[1]=a[1] a[2]=a[2] a[3]=a[3] c[0]=w1[0] c[1]=w1[1] c[2]=w1[2] c[3]=w1[3] pclk=clk
.subckt passthrough a[0]=w1[0] a[1]=w1[1] a[2]=w1[2] a[3]=w1[3] c[0]=w2[0] c[1]=w2[1] c[2]=w2[2] c[3]=w2[3] pclk=clk
.end

.model my_4bit_dff
.inputs mclk din[0] din[1] din[2] din[3] en
.outputs q[0] q[1] q[2] q[3]
.names $false
.names $true
1
.names $undef
.names q[0] din[0] en $0\q[3:0][0]
1-0 1
-11 1
.names q[1] din[1] en $0\q[3:0][1]
1-0 1
-11 1
.names q[2] din[2] en $0\q[3:0][2]
1-0 1
-11 1
.names q[3] din[3] en $0\q[3:0][3]
1-0 1
-11 1
.latch $0\q[3:0][0] q[0] re mclk 2
.latch $0\q[3:0][1] q[1] re mclk 2
.latch $0\q[3:0][2] q[2] re mclk 2
.latch $0\q[3:0][3] q[3] re mclk 2
.end

.model passthrough
.inputs a[0] a[1] a[2] a[3] pclk
.outputs c[0] c[1] c[2] c[3]
.names $false
.names $true
1
.names $undef
.subckt my_4bit_dff din[0]=a[0] din[1]=a[1] din[2]=a[2] din[3]=a[3] en=$true mclk=pclk q[0]=c[0] q[1]=c[1] q[2]=c[2] q[3]=c[3]
.end
"""  # noqa


class TestInputFromBlif(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_combo_blif_input_has_correct_io_interface(self):
        pyrtl.input_from_blif(full_adder_blif)
        x, y, cin, sumw, cout, bad = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["x", "y", "cin", "sum", "cout", "bad"]
        ]
        self.assertIsNotNone(x)
        self.assertIsNotNone(y)
        self.assertIsNotNone(cin)
        self.assertIsNotNone(sumw)
        self.assertIsNotNone(cout)
        self.assertIsNone(bad)
        self.assertEqual(len(x), 1)
        self.assertEqual(len(y), 1)
        self.assertEqual(len(cin), 1)
        self.assertEqual(len(sumw), 1)
        self.assertEqual(len(cout), 1)
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertIn(x, io_input)
        self.assertIn(y, io_input)
        self.assertIn(cin, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertIn(sumw, io_output)
        self.assertIn(cout, io_output)

    def test_sequential_blif_input_has_correct_io_interface(self):
        pyrtl.input_from_blif(state_machine_blif)
        inw, reset, out = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["in", "reset", "out"]
        ]
        self.assertIsNotNone(inw)
        self.assertIsNotNone(reset)
        self.assertIsNotNone(out)
        self.assertEqual(len(inw), 1)
        self.assertEqual(len(reset), 1)
        self.assertEqual(len(out), 4)
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertIn(inw, io_input)
        self.assertIn(reset, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertIn(out, io_output)

    def test_sequential_blif_input_has_correct_io_interface_counter(self):
        pyrtl.input_from_blif(counter4bit_blif)
        rst, en, count = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["rst", "en", "count"]
        ]
        self.assertIsNotNone(rst)
        self.assertIsNotNone(en)
        self.assertIsNotNone(count)
        self.assertEqual(len(rst), 1)
        self.assertEqual(len(en), 1)
        self.assertEqual(len(count), 4)
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertIn(rst, io_input)
        self.assertIn(en, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertIn(count, io_output)

    def test_correct_interface_with_unmerged_io(self):
        pyrtl.input_from_blif(simple_unmerged_io_blif, merge_io_vectors=False)
        a0, a1, a2, a3, b0, b1 = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["a[0]", "a[1]", "a[2]", "a[3]", "b[0]", "b[1]"]
        ]
        self.assertEqual(len(a0), 1)
        self.assertEqual(len(a1), 1)
        self.assertEqual(len(a2), 1)
        self.assertEqual(len(a3), 1)
        self.assertEqual(len(b0), 1)
        self.assertEqual(len(b1), 1)
        self.assertEqual(
            {a0, a1, a2, a3}, pyrtl.working_block().wirevector_subset(pyrtl.Input)
        )
        self.assertEqual(
            {b0, b1}, pyrtl.working_block().wirevector_subset(pyrtl.Output)
        )

    def test_blif_input_simulates_correctly_with_merged_outputs(self):
        # The 'counter_blif' string contains a model of a standard 4-bit
        # synchronous-reset counter with enable. In particular, the model has 4 1-bit
        # outputs named "count[0]", "count[1]", "count[2]", and "count[3]". The internal
        # PyRTL representation will by default convert these related 1-bit wires into a
        # single 4-bit wire called "count". This test simulates the design and, among
        # other things, ensures that this output wire conversion occurred correctly.
        pyrtl.input_from_blif(counter4bit_blif)
        io_vectors = pyrtl.working_block().wirevector_subset(
            (pyrtl.Input, pyrtl.Output)
        )
        sim = pyrtl.Simulation(tracer=pyrtl.SimulationTrace(wires_to_track=io_vectors))
        inputs = {
            "rst": [1] + [0] * 20,
            "en": [1] + [1] * 20,
        }
        expected = {"count": [0, *list(range(0, 16)), *list(range(0, 4))]}
        sim.step_multiple(inputs, expected)

        correct_output = (
            "  --- Values in base 10 ---\n"
            "count  0  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15  0  1  2  3\n"
            "en     1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1\n"
            "rst    1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_blif_input_simulates_correctly_with_unmerged_outputs(self):
        pyrtl.input_from_blif(counter4bit_blif, merge_io_vectors=False)
        count0, count1, count2, count3 = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["count[0]", "count[1]", "count[2]", "count[3]"]
        ]
        self.assertEqual(len(count0), 1)
        self.assertEqual(len(count1), 1)
        self.assertEqual(len(count2), 1)
        self.assertEqual(len(count3), 1)
        io_vectors = pyrtl.working_block().wirevector_subset(
            (pyrtl.Input, pyrtl.Output)
        )
        sim = pyrtl.Simulation(tracer=pyrtl.SimulationTrace(wires_to_track=io_vectors))
        inputs = {
            "rst": [1] + [0] * 20,
            "en": [1] + [1] * 20,
        }
        expected_merged = [0, *list(range(0, 16)), *list(range(0, 4))]

        expected = {
            "count[0]": [n & 0b0001 for n in expected_merged],
            "count[1]": [(n & 0b0010) >> 1 for n in expected_merged],
            "count[2]": [(n & 0b0100) >> 2 for n in expected_merged],
            "count[3]": [(n & 0b1000) >> 3 for n in expected_merged],
        }
        sim.step_multiple(inputs, expected)

        correct_output = (
            "     --- Values in base 10 ---\n"
            "count[0] 0 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1\n"
            "count[1] 0 0 0 1 1 0 0 1 1 0 0 1 1 0 0 1 1 0 0 1 1\n"
            "count[2] 0 0 0 0 0 1 1 1 1 0 0 0 0 1 1 1 1 0 0 0 0\n"
            "count[3] 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0\n"
            "en       1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1\n"
            "rst      1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_blif_with_output_as_arg(self):
        pyrtl.input_from_blif(blif_with_output_as_arg)
        inw, outw = [
            pyrtl.working_block().get_wirevector_by_name(s) for s in ["in", "out"]
        ]
        self.assertIsNotNone(inw)
        self.assertIsNotNone(outw)
        self.assertEqual(len(inw), 2)
        self.assertEqual(len(outw), 1)
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertIn(inw, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertIn(outw, io_output)

    def test_blif_with_multiple_modules_merged_io(self):
        pyrtl.input_from_blif(four_bit_adder_multi_module)
        a, b, cin, s, cout = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in ["a", "b", "cin", "s", "cout"]
        ]
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertIn(a, io_input)
        self.assertIn(b, io_input)
        self.assertIn(cin, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertIn(s, io_output)
        self.assertIn(cout, io_output)

        self.assertEqual(len(a), 4)
        self.assertEqual(len(b), 4)
        self.assertEqual(len(cin), 1)
        self.assertEqual(len(s), 4)
        self.assertEqual(len(cout), 1)

        avals = list(range(0, 16)) * 2
        bvals = list(range(0, 16)) * 2
        cinvals = [0] * 16 + [1] * 16

        res = utils.sim_and_ret_outws([a, b, cin], [avals, bvals, cinvals])
        self.assertEqual(
            res[s.name],
            [(av + bv + cinv) & 0xF for av, bv, cinv in zip(avals, bvals, cinvals)],
        )
        self.assertEqual(
            res[cout.name],
            [
                ((av + bv + cinv) & 0x10) >> 4
                for av, bv, cinv in zip(avals, bvals, cinvals)
            ],
        )

    def test_blif_with_multiple_modules_unmerged_io(self):
        pyrtl.input_from_blif(four_bit_adder_multi_module, merge_io_vectors=False)
        a0, a1, a2, a3, b0, b1, b2, b3, cin, s0, s1, s2, s3, cout = [
            pyrtl.working_block().get_wirevector_by_name(s)
            for s in [
                "a[0]",
                "a[1]",
                "a[2]",
                "a[3]",
                "b[0]",
                "b[1]",
                "b[2]",
                "b[3]",
                "cin",
                "s[0]",
                "s[1]",
                "s[2]",
                "s[3]",
                "cout",
            ]
        ]
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertEqual({a0, a1, a2, a3, b0, b1, b2, b3, cin}, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertEqual({s0, s1, s2, s3, cout}, io_output)

        self.assertEqual(len(a0), 1)
        self.assertEqual(len(a1), 1)
        self.assertEqual(len(a2), 1)
        self.assertEqual(len(a3), 1)
        self.assertEqual(len(b0), 1)
        self.assertEqual(len(b1), 1)
        self.assertEqual(len(b2), 1)
        self.assertEqual(len(b3), 1)
        self.assertEqual(len(cin), 1)
        self.assertEqual(len(s0), 1)
        self.assertEqual(len(s1), 1)
        self.assertEqual(len(s2), 1)
        self.assertEqual(len(s3), 1)
        self.assertEqual(len(cout), 1)

        sim = pyrtl.Simulation()
        for a in range(0, 16):
            for b in range(0, 16):
                for cin in range(0, 1):
                    sim.step(
                        {
                            "a[0]": a & 0x1,
                            "a[1]": (a & 0x2) >> 1,
                            "a[2]": (a & 0x4) >> 2,
                            "a[3]": (a & 0x8) >> 3,
                            "b[0]": b & 0x1,
                            "b[1]": (b & 0x2) >> 1,
                            "b[2]": (b & 0x4) >> 2,
                            "b[3]": (b & 0x8) >> 3,
                            "cin": cin,
                        }
                    )
                    res = a + b + cin
                    self.assertEqual(sim.inspect("s[0]"), res & 0x1)
                    self.assertEqual(sim.inspect("s[1]"), (res & 0x2) >> 1)
                    self.assertEqual(sim.inspect("s[2]"), (res & 0x4) >> 2)
                    self.assertEqual(sim.inspect("s[3]"), (res & 0x8) >> 3)
                    self.assertEqual(sim.inspect("cout"), (res & 0x10) >> 4)

    def test_blif_with_clock_passing(self):
        pyrtl.input_from_blif(clock_passing_blif)
        a, b, c = [
            pyrtl.working_block().get_wirevector_by_name(s) for s in ["a", "b", "c"]
        ]
        io_input = pyrtl.working_block().wirevector_subset(pyrtl.Input)
        self.assertEqual({a, b}, io_input)
        io_output = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        self.assertEqual({c}, io_output)

        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": [0, 3, 1, 1, 1, 1, 0],
                "b": [1, 7, 9, 9, 9, 9, 0],
            }
        )
        cvals = sim.tracer.trace[c.name]
        self.assertEqual(cvals, [0, 0, 0, 1, 1, 1, 0])

    def test_blif_error_zeroes_in_offset(self):
        zeroes_in_offset = """\
        .model Top
        .inputs clk in[0] in[1]
        .outputs out
        .names in[0] in[1] out
        10 0
        .end
        """

        with self.assertRaisesRegex(pyrtl.PyrtlError, "Off-set found"):
            pyrtl.input_from_blif(zeroes_in_offset)

    def test_blif_error_bad_coverset(self):
        bad_coverset = """\
        .model Top
        .inputs clk in[0] in[1]
        .outputs out
        .names in[0] in[1] out
        10 1 1
        .end
        """
        with self.assertRaisesRegex(pyrtl.PyrtlError, "malformed cover set"):
            pyrtl.input_from_blif(bad_coverset)

    def test_blif_not_gate_correct(self):
        blif = """\
        .model Top
        .inputs a
        .outputs o
        .names a o
        0 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("~")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "01",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [1, 0])

    def test_blif_and_gate_correct(self):
        blif = """\
        .model Top
        .inputs a b
        .outputs o
        .names a b o
        11 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("&")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "0011",
                "b": "0101",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [0, 0, 0, 1])

    def test_blif_or_gate_correct(self):
        blif = """\
        .model Top
        .inputs a b
        .outputs o
        .names a b o
        1- 1
        -1 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("|")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "0011",
                "b": "0101",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [0, 1, 1, 1])

    def test_blif_nand_gate_to_primitives_correct(self):
        # This tests that there should be no NAND gates generated during BLIF import;
        # they should be converted to AND+NOT.
        blif = """\
        .model Top
        .inputs a b
        .outputs o
        .names a b o
        0- 1
        -0 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("n")), 0)
        self.assertEqual(len(block.logic_subset("&")), 1)
        self.assertEqual(len(block.logic_subset("~")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "0011",
                "b": "0101",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [1, 1, 1, 0])

    def test_blif_xor_gate_correct(self):
        blif = """\
        .model Top
        .inputs a b
        .outputs o
        .names a b o
        10 1
        01 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("^")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "0011",
                "b": "0101",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [0, 1, 1, 0])

    def test_blif_nor_gate_correct(self):
        # This is a non-primitive, so tests the last branch of cover list parsing
        blif = """\
        .model Top
        .inputs a b
        .outputs o
        .names a b o
        00 1
        .end
        """
        pyrtl.input_from_blif(blif)
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset("~")), 2)
        self.assertEqual(len(block.logic_subset("&")), 1)
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": "0011",
                "b": "0101",
            }
        )
        self.assertEqual(sim.tracer.trace["o"], [1, 0, 0, 0])


verilog_output_small = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, o);
    input clk;
    output[12:0] o;

    // Constants
    wire[5:0] k = 6'd38;

    // Combinational logic
    assign o = {4'd12, 3'd3, k};
endmodule
"""


verilog_output_large = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, rst, a, o);
    input clk;
    input rst;
    input[3:0] a;
    output[5:0] o;

    // Memories
    reg[3:0] z[3:0];  // MemBlock
    reg[3:0] tmp0[3:0];  // MemBlock
    reg[3:0] tmp1[3:0];  // MemBlock
    reg[3:0] tmp2[3:0];  // MemBlock
    reg[3:0] tmp3[3:0];  // MemBlock
    reg[3:0] tmp4[3:0];  // MemBlock
    reg[3:0] tmp5[3:0];  // MemBlock
    reg[3:0] tmp6[3:0];  // MemBlock
    reg[3:0] tmp7[3:0];  // MemBlock
    reg[3:0] tmp8[3:0];  // MemBlock
    reg[3:0] tmp9[3:0];  // MemBlock
    reg[3:0] tmp10[3:0];  // MemBlock
    reg[3:0] tmp11[3:0];  // MemBlock

    // Registers
    reg[3:0] r;
    reg[3:0] s;

    // Temporaries
    wire[4:0] tmp14;
    wire[4:0] tmp18;
    wire[4:0] tmp22;
    wire[4:0] tmp26;
    wire[4:0] tmp30;
    wire[4:0] tmp34;
    wire[4:0] tmp38;
    wire[4:0] tmp42;
    wire[4:0] tmp46;
    wire[4:0] tmp50;
    wire[4:0] tmp54;
    wire[4:0] tmp58;
    wire[4:0] tmp60;
    wire[6:0] tmp66;
    wire[4:0] tmp70;
    wire[3:0] tmp72;
    wire[3:0] tmp76;
    wire[6:0] tmp79;

    // Combinational logic
    assign o = (tmp79[5:0]);
    assign tmp14 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp18 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp22 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp26 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp30 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp34 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp38 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp42 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp46 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp50 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp54 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp58 = (r + {{3 {1'd0}}, 1'd1});
    assign tmp60 = (a + r);
    assign tmp66 = ((tmp60 + {{4 {1'd0}}, 1'd1}) - {{2 {1'd0}}, s});
    assign tmp70 = (a - {{3 {1'd0}}, 1'd1});
    assign tmp79 = ((tmp60 + {(1'd0), tmp72}) + {{2 {1'd0}}, tmp76});

    // Register logic
    always @(posedge clk) begin
        if (rst) begin
            r <= 4'd0;
            s <= 4'd13;
        end else begin
            r <= (tmp66[3:0]);
            s <= (tmp70[3:0]);
        end
    end

    // MemBlock z logic
    always @(posedge clk) begin
        z[2'd0] <= 4'd9;
    end

    // MemBlock tmp0 logic
    always @(posedge clk) begin
        tmp0[2'd0] <= a;
        tmp0[2'd1] <= (tmp14[3:0]);
    end
    assign tmp72 = tmp0[2'd0];

    // MemBlock tmp1 logic
    always @(posedge clk) begin
        tmp1[2'd0] <= a;
        tmp1[2'd1] <= (tmp18[3:0]);
    end
    assign tmp76 = tmp1[2'd0];

    // MemBlock tmp2 logic
    always @(posedge clk) begin
        tmp2[2'd0] <= a;
        tmp2[2'd1] <= (tmp22[3:0]);
    end

    // MemBlock tmp3 logic
    always @(posedge clk) begin
        tmp3[2'd0] <= a;
        tmp3[2'd1] <= (tmp26[3:0]);
    end

    // MemBlock tmp4 logic
    always @(posedge clk) begin
        tmp4[2'd0] <= a;
        tmp4[2'd1] <= (tmp30[3:0]);
    end

    // MemBlock tmp5 logic
    always @(posedge clk) begin
        tmp5[2'd0] <= a;
        tmp5[2'd1] <= (tmp34[3:0]);
    end

    // MemBlock tmp6 logic
    always @(posedge clk) begin
        tmp6[2'd0] <= a;
        tmp6[2'd1] <= (tmp38[3:0]);
    end

    // MemBlock tmp7 logic
    always @(posedge clk) begin
        tmp7[2'd0] <= a;
        tmp7[2'd1] <= (tmp42[3:0]);
    end

    // MemBlock tmp8 logic
    always @(posedge clk) begin
        tmp8[2'd0] <= a;
        tmp8[2'd1] <= (tmp46[3:0]);
    end

    // MemBlock tmp9 logic
    always @(posedge clk) begin
        tmp9[2'd0] <= a;
        tmp9[2'd1] <= (tmp50[3:0]);
    end

    // MemBlock tmp10 logic
    always @(posedge clk) begin
        tmp10[2'd0] <= a;
        tmp10[2'd1] <= (tmp54[3:0]);
    end

    // MemBlock tmp11 logic
    always @(posedge clk) begin
        tmp11[2'd0] <= a;
        tmp11[2'd1] <= (tmp58[3:0]);
    end
endmodule
"""


verilog_output_mems_with_no_writes = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, rst, in1, out1);
    input clk;
    input rst;
    input[2:0] in1;
    output[7:0] out1;

    // Memories
    reg[7:0] rom[7:0];  // RomBlock
    reg[7:0] mem[255:0];  // MemBlock

    // Temporaries
    wire[7:0] tmp0;

    // Read-only memory data
    initial begin
        rom[0] = 8'ha;
        rom[1] = 8'h14;
        rom[2] = 8'h1e;
        rom[3] = 8'h28;
        rom[4] = 8'h32;
        rom[5] = 8'h3c;
        rom[6] = 8'h0;
        rom[7] = 8'h0;
    end

    // Combinational logic
    assign out1 = tmp0;

    // RomBlock rom logic
    assign tmp0 = rom[in1];

    // MemBlock mem logic
    always @(posedge clk) begin
        mem[tmp0] <= 8'd42;
    end
endmodule
"""


verilog_output_counter_sync_reset = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, rst, o);
    input clk;
    input rst;
    output[3:0] o;

    // Registers
    reg[3:0] tmp0;

    // Temporaries
    wire[4:0] tmp3;

    // Combinational logic
    assign o = tmp0;
    assign tmp3 = (tmp0 + {{3 {1'd0}}, 1'd1});

    // Register logic
    always @(posedge clk) begin
        if (rst) begin
            tmp0 <= 4'd2;
        end else begin
            tmp0 <= (tmp3[3:0]);
        end
    end
endmodule
"""


verilog_output_counter_async_reset = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, rst, o);
    input clk;
    input rst;
    output[3:0] o;

    // Registers
    reg[3:0] tmp0;

    // Temporaries
    wire[4:0] tmp3;

    // Combinational logic
    assign o = tmp0;
    assign tmp3 = (tmp0 + {{3 {1'd0}}, 1'd1});

    // Register logic
    always @(posedge clk or posedge rst) begin
        if (rst) begin
            tmp0 <= 4'd2;
        end else begin
            tmp0 <= (tmp3[3:0]);
        end
    end
endmodule
"""


verilog_output_counter_no_reset = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, o);
    input clk;
    output[3:0] o;

    // Registers
    reg[3:0] tmp0;

    // Temporaries
    wire[4:0] tmp3;

    // Combinational logic
    assign o = tmp0;
    assign tmp3 = (tmp0 + {{3 {1'd0}}, 1'd1});

    // Register logic
    always @(posedge clk) begin
        tmp0 <= (tmp3[3:0]);
    end
endmodule
"""


verilog_custom_reset = """\
// Generated automatically via PyRTL
// As one initial test of synthesis, map to FPGA with:
//   yosys -p "synth_xilinx -top toplevel" thisfile.v

module toplevel(clk, rst);
    input clk;
    input rst;

    // Registers
    reg[3:0] r;

    // Temporaries
    wire[4:0] tmp5;

    // Combinational logic
    assign tmp5 = (rst ? {{4 {1'd0}}, 1'd0} : (r + {{3 {1'd0}}, 1'd1}));

    // Register logic
    always @(posedge clk) begin
        r <= (tmp5[3:0]);
    end
endmodule
"""


class TestVerilogNames(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.vnames = _VerilogSanitizer("_sani_test")

    def checkname(self, name):
        self.assertEqual(self.vnames.make_valid_string(name), name)

    def assert_invalid_name(self, name):
        self.assertNotEqual(self.vnames.make_valid_string(name), name)

    def test_verilog_check_valid_name_good(self):
        self.checkname("abc")
        self.checkname("a")
        self.checkname("BC")
        self.checkname("Kabc")
        self.checkname("B_ac")
        self.checkname("_asdvqa")
        self.checkname("_Bs_")
        self.checkname("fd$oeoe")
        self.checkname("_B$$s")
        self.checkname("B")

    def test_verilog_check_valid_name_bad(self):
        self.assert_invalid_name("carne asda")
        self.assert_invalid_name("")
        self.assert_invalid_name("asd%kask")
        self.assert_invalid_name("flipin'")
        self.assert_invalid_name(" jklol")
        self.assert_invalid_name("a" * 2000)


class TestVerilogOutput(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        # To compare textual consistency, need to make sure we're starting at the same
        # index for all automatically created names.
        pyrtl.wire._reset_wire_indexers()
        pyrtl.memory._reset_memory_indexer()

        self.maxDiff = 30000

    def test_romblock_does_not_throw_error(self):
        a = pyrtl.Input(bitwidth=3, name="a")
        b = pyrtl.Input(bitwidth=3, name="b")
        o = pyrtl.Output(bitwidth=3, name="o")
        mixtable = pyrtl.RomBlock(
            addrwidth=3,
            bitwidth=3,
            pad_with_zeros=True,
            romdata={0: 1, 1: 2, 2: 5, 5: 0},
            asynchronous=True,
        )
        res = _basic_add(a, b)
        o <<= mixtable[res[:-1]]
        with io.StringIO() as testbuffer:
            pyrtl.output_to_verilog(testbuffer)

    def test_textual_consistency_small(self):
        i = pyrtl.Const(12)
        j = pyrtl.Const(3, bitwidth=3)
        k = pyrtl.Const(38, name="k")
        o = pyrtl.Output(bitwidth=13, name="o")
        o <<= pyrtl.concat(i, j, k)

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer, add_reset=False)

        self.assertEqual(buffer.getvalue(), verilog_output_small)

    def test_textual_consistency_large(self):
        # The following is a non-sensical program created to test that the Verilog that
        # is created is deterministic in the order in which it presents the wire,
        # register, and memory declarations and the combinational and sequential logic.
        # Hence it creates many memories, and makes sure at least two lines of code are
        # created in the always @ blocks associated with them (so we have many different
        # wire names to deal with and test against).
        a = pyrtl.Input(bitwidth=4, name="a")
        r = pyrtl.Register(bitwidth=4, name="r")
        s = pyrtl.Register(bitwidth=4, name="s", reset_value=13)
        # This will have mem id 0, so prints first despite actual name
        mt = pyrtl.MemBlock(bitwidth=4, addrwidth=2, name="z")
        m = [
            pyrtl.MemBlock(bitwidth=4, addrwidth=2, max_write_ports=2)
            for _ in range(12)
        ]
        for mem in m:
            mem[0] <<= a
            mem[1] <<= (r + 1).truncate(4)
        b = a + r
        r.next <<= b + 1 - s
        s.next <<= a - 1
        mt[0] <<= 9
        o = pyrtl.Output(6, "o")
        o <<= b + m[0][0] + m[1][0]

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer)

        self.assertEqual(buffer.getvalue(), verilog_output_large)

    def test_mems_with_no_writes(self):
        rdata = {0: 10, 1: 20, 2: 30, 3: 40, 4: 50, 5: 60}
        rom = pyrtl.RomBlock(
            name="rom", bitwidth=8, addrwidth=3, romdata=rdata, pad_with_zeros=True
        )
        mem = pyrtl.MemBlock(name="mem", bitwidth=8, addrwidth=8)
        in1 = pyrtl.Input(bitwidth=3, name="in1")
        out1 = pyrtl.Output(bitwidth=8, name="out1")
        w = rom[in1]
        out1 <<= w
        mem[w] <<= 42

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer)

        self.assertEqual(buffer.getvalue(), verilog_output_mems_with_no_writes)

    def check_counter_text(self, add_reset, expected):
        r = pyrtl.Register(bitwidth=4, reset_value=2)
        r.next <<= r + 1
        o = pyrtl.Output(bitwidth=4, name="o")
        o <<= r

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer, add_reset)
        self.assertEqual(buffer.getvalue(), expected)

    def test_textual_consistency_with_sync_reset(self):
        self.check_counter_text(True, verilog_output_counter_sync_reset)

    def test_textual_consistency_with_async_reset(self):
        self.check_counter_text("asynchronous", verilog_output_counter_async_reset)

    def test_textual_consistency_with_no_reset(self):
        self.check_counter_text(False, verilog_output_counter_no_reset)

    def test_error_invalid_add_reset(self):
        buffer = io.StringIO()
        with self.assertRaisesRegex(pyrtl.PyrtlError, "Invalid add_reset option"):
            pyrtl.output_to_verilog(buffer, add_reset="foobar")

    def test_error_existing_reset_wire(self):
        buffer = io.StringIO()
        _rst = pyrtl.Input(bitwidth=1, name="rst")
        with self.assertRaisesRegex(
            pyrtl.PyrtlError, "Found a user-defined wire named 'rst'."
        ):
            pyrtl.output_to_verilog(buffer)

    def test_existing_reset_wire_without_add_reset(self):
        buffer = io.StringIO()
        rst = pyrtl.Input(bitwidth=1, name="rst")
        r = pyrtl.Register(bitwidth=4, name="r")
        r.next <<= pyrtl.select(rst, 0, r + 1)
        pyrtl.output_to_verilog(buffer, add_reset=False)
        self.assertEqual(buffer.getvalue(), verilog_custom_reset)

    def test_register_reset_value(self):
        register0 = pyrtl.Register(name="register0", bitwidth=8, reset_value=0)
        register1 = pyrtl.Register(name="register1", bitwidth=4, reset_value=1)
        register0.next <<= 0
        register1.next <<= 1

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer, add_reset=False, initialize_registers=True)

        self.assertTrue("reg[7:0] register0 = 8'd0" in buffer.getvalue())
        self.assertTrue("reg[3:0] register1 = 4'd1" in buffer.getvalue())

    def test_bit_slice_inputs(self):
        """Verify that wires are always declared for bit-slice inputs, even Consts."""
        a = pyrtl.Input(name="a", bitwidth=1)
        b = pyrtl.Input(name="b", bitwidth=1)
        c = pyrtl.Input(name="c", bitwidth=2)

        x = pyrtl.Output(name="x", bitwidth=1)
        y = pyrtl.Output(name="y", bitwidth=1)
        z = pyrtl.Output(name="z", bitwidth=1)

        x <<= pyrtl.Const(42)[1]

        y <<= (a + b)[1]

        t = pyrtl.WireVector()
        t <<= c
        z <<= t[1]

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer)

        # A constant should be declared for ``42``, even though it has no user-specified
        # name and only has one user, because that user is a bit-slice.
        self.assertTrue("wire[5:0] const_0_42 = 6'd42" in buffer.getvalue())
        # A temporary wire should be declared for ``a + b``, even though it has no name
        # and only has one user, because that user is a bit-slice.
        self.assertTrue("assign tmp1 = (a + b)" in buffer.getvalue())
        # A temporary wire should be declared for ``t``, even though it has no name and
        # only has one user and just passes through ``c``, because that user is a
        # bit-slice.
        self.assertTrue("assign tmp3 = c" in buffer.getvalue())


verilog_input_counter = """\
module counter (clk, rst, en, count);

    input clk, rst, en;
    output reg [3:0] count;

    always @(posedge clk)
        if (rst)
            count <= 4'd0;
        else if (en)
            count <= count + 4'd1;

endmodule
"""

verilog_input_multi_module = """\
module foo (a, b, o);

    input a, b;
    output [1:0] o;
    assign o = a + b;

endmodule

module top (clk, o);
    input clk;
    reg a, b;
    output [1:0] o;
    foo f1(a, b, o);

    always @(posedge clk) begin
        a <= ~a;
        b <= ~b;
    end
endmodule
"""


class TestVerilogInput(unittest.TestCase):
    def setUp(self):
        try:
            _ = subprocess.check_output(["yosys", "-V"])
        except OSError as exc:
            msg = "Testing Verilog input requires yosys"
            raise unittest.SkipTest(msg) from exc
        pyrtl.reset_working_block()

    def test_import_counter(self):
        pyrtl.input_from_verilog(verilog_input_counter)
        sim = pyrtl.Simulation()
        sim.step_multiple({"rst": "10000", "en": "01111"})
        self.assertEqual(sim.tracer.trace["count"], [0, 0, 1, 2, 3])

    def test_import_small(self):
        pyrtl.input_from_verilog(verilog_output_small)
        sim = pyrtl.Simulation()
        sim.step()
        self.assertEqual(sim.tracer.trace["o"][0], 0b1100011100110)

    def test_import_counter_with_reset(self):
        pyrtl.input_from_verilog(verilog_output_counter_sync_reset)
        sim = pyrtl.Simulation()
        sim.step_multiple({"rst": "1000"})
        self.assertEqual(sim.tracer.trace["o"], [0, 2, 3, 4])

    def test_import_multi_module_specified_module(self):
        # Import foo module because occurs first in file
        pyrtl.input_from_verilog(verilog_input_multi_module, toplevel="foo")
        sim = pyrtl.Simulation()
        sim.step_multiple({"a": "0011", "b": "0101"})
        self.assertEqual(sim.tracer.trace["o"], [0, 1, 1, 2])

    def test_import_multi_module_auto_select_top_module(self):
        pyrtl.input_from_verilog(verilog_input_multi_module)
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=5)
        self.assertEqual(sim.tracer.trace["o"], [0, 2, 0, 2, 0])

    def test_error_import_bad_file(self):
        with self.assertRaisesRegex(
            pyrtl.PyrtlError, "input_from_verilog expecting either open file or string"
        ):
            pyrtl.input_from_verilog(3)


verilog_testbench = """\
module tb();
    reg clk;
    reg rst;

    // block Inputs
    reg[1:0] a100;
    reg[3:0] w1;
    reg[2:0] w12;

    // block Outputs
    wire[1:0] out1;
    wire[8:0] out10;

    integer tb_addr;
    toplevel block(.clk(clk), .rst(rst), .a100(a100), .w1(w1), .w12(w12), .out1(out1), .out10(out10));

    always
        #5 clk = ~clk;

    initial begin
        $dumpfile ("waveform.vcd");
        $dumpvars;

        clk = 1'd0;
        rst = 1'd0;

        // Initialize Registers
        block.r1 = 3'd2;
        block.r2 = 4'd3;
        block.tmp0 = 8'd0;

        // Initialize MemBlocks
        for (tb_addr = 0; tb_addr < 32; tb_addr++) begin block.mem[tb_addr] = 4'd0; end
        block.mem[2] = 4'd9;
        block.mem[9] = 4'd12;
        a100 = 2'd0;
        w1 = 4'd0;
        w12 = 3'd0;

        #10
        a100 = 2'd1;
        w1 = 4'd4;
        w12 = 3'd1;

        #10
        a100 = 2'd3;
        w1 = 4'd2;
        w12 = 3'd7;

        #10
        a100 = 2'd2;
        w1 = 4'd3;
        w12 = 3'd4;

        #10
        $finish;
    end
endmodule
"""  # noqa

verilog_testbench_no_reset = """\
module tb();
    reg clk;

    // block Inputs
    reg[1:0] a100;
    reg[3:0] w1;
    reg[2:0] w12;

    // block Outputs
    wire[1:0] out1;
    wire[8:0] out10;

    integer tb_addr;
    toplevel block(.clk(clk), .a100(a100), .w1(w1), .w12(w12), .out1(out1), .out10(out10));

    always
        #5 clk = ~clk;

    initial begin
        $dumpfile ("waveform.vcd");
        $dumpvars;

        clk = 1'd0;

        // Initialize Registers
        block.r1 = 3'd2;
        block.r2 = 4'd3;
        block.tmp0 = 8'd0;

        // Initialize MemBlocks
        for (tb_addr = 0; tb_addr < 32; tb_addr++) begin block.mem[tb_addr] = 4'd0; end
        block.mem[2] = 4'd9;
        block.mem[9] = 4'd12;
        a100 = 2'd0;
        w1 = 4'd0;
        w12 = 3'd0;

        #10
        a100 = 2'd1;
        w1 = 4'd4;
        w12 = 3'd1;

        #10
        a100 = 2'd3;
        w1 = 4'd2;
        w12 = 3'd7;

        #10
        a100 = 2'd2;
        w1 = 4'd3;
        w12 = 3'd4;

        #10
        $finish;
    end
endmodule
"""  # noqa: E501

verilog_testbench_custom_reset = """\
module tb();
    reg clk;

    // block Inputs
    reg rst;

    toplevel block(.clk(clk), .rst(rst));

    always
        #5 clk = ~clk;

    initial begin
        $dumpfile ("waveform.vcd");
        $dumpvars;

        clk = 1'd0;

        // Initialize Registers
        block.r = 4'd0;
        $finish;
    end
endmodule
"""


class TestOutputTestbench(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        # To compare textual consistency, need to make sure we're starting at the same
        # index for all automatically created names.
        pyrtl.wire._reset_wire_indexers()
        pyrtl.memory._reset_memory_indexer()
        self.maxDiff = 10000

    def test_verilog_testbench_does_not_throw_error(self):
        zero = pyrtl.Input(1, "zero")
        counter_output = pyrtl.Output(3, "counter_output")
        counter = pyrtl.Register(3, "counter")
        counter.next <<= pyrtl.mux(zero, counter + 1, 0)
        counter_output <<= counter
        sim = pyrtl.Simulation(tracer=pyrtl.SimulationTrace([counter_output, zero]))
        for _cycle in range(15):
            sim.step({zero: random.choice([0, 0, 0, 1])})
        with io.StringIO() as tbfile:
            pyrtl.output_verilog_testbench(tbfile, sim.tracer)

    def create_design(self):
        # Various wire names so we can verify they are printed in deterministic order
        # each time
        i1, i2, i3 = pyrtl.input_list("w1/4 w12/3 a100/2")
        r1, r2 = pyrtl.register_list("r1/3 r2/4")
        r3 = pyrtl.Register(bitwidth=8)
        mem = pyrtl.MemBlock(name="mem", bitwidth=4, addrwidth=5)
        o1, o2 = pyrtl.output_list("out1/2 out10/9")
        r1.next <<= i1 + i2
        r2.next <<= r1 * i3
        r3.next <<= r1 & r2
        mem[i1] <<= r1 + 3
        o1 <<= i3 - r2
        o2 <<= r1
        sim = pyrtl.Simulation(
            register_value_map={
                r1: 2,
                r2: 3,
            },
            memory_value_map={
                mem: {
                    2: 9,
                    9: 12,
                },
            },
        )
        sim.step_multiple(
            {
                "w1": [0, 4, 2, 3],
                "w12": [0, 1, 7, 4],
                "a100": [0, 1, 3, 2],
            }
        )
        return sim.tracer

    def test_verilog_testbench_consistency(self):
        sim_trace = self.create_design()
        with io.StringIO() as tbfile:
            pyrtl.output_verilog_testbench(tbfile, sim_trace)
            self.assertEqual(tbfile.getvalue(), verilog_testbench)

    def test_verilog_testbench_no_reset_consistency(self):
        sim_trace = self.create_design()
        with io.StringIO() as tbfile:
            pyrtl.output_verilog_testbench(tbfile, sim_trace, add_reset=False)
            self.assertEqual(tbfile.getvalue(), verilog_testbench_no_reset)

    def test_error_verilog_testbench_invalid_add_reset(self):
        tbfile = io.StringIO()
        with self.assertRaisesRegex(pyrtl.PyrtlError, "Invalid add_reset option"):
            pyrtl.output_verilog_testbench(tbfile, add_reset="foobar")

    def test_error_verilog_testbench_existing_reset_wire(self):
        tbfile = io.StringIO()
        _rst = pyrtl.Input(1, "rst")
        with self.assertRaisesRegex(
            pyrtl.PyrtlError, "Found a user-defined wire named 'rst'."
        ):
            pyrtl.output_verilog_testbench(tbfile)

    def test_verilog_testbench_existing_reset_wire_without_add_reset(self):
        rst = pyrtl.Input(1, "rst")
        r = pyrtl.Register(4, "r")
        r.next <<= pyrtl.select(rst, 0, r + 1)

        buffer = io.StringIO()
        pyrtl.output_verilog_testbench(buffer, add_reset=False)
        self.assertEqual(buffer.getvalue(), verilog_testbench_custom_reset)

    def test_only_initialize_memblocks(self):
        """Test that RomBlocks are not re-initialized by the testbench."""
        romblock = pyrtl.RomBlock(
            name="my_rom", bitwidth=3, addrwidth=2, romdata=[1, 2, 3, 4]
        )
        addr = pyrtl.Input(name="addr", bitwidth=2)
        data = pyrtl.Output(name="data", bitwidth=3)
        data <<= romblock[addr]

        buffer = io.StringIO()
        pyrtl.output_to_verilog(buffer)
        # The Verilog code should declare the RomBlock and define its values.
        self.assertTrue("reg[2:0] my_rom[3:0];  // RomBlock" in buffer.getvalue())
        self.assertTrue("my_rom[0] = 3'h1;" in buffer.getvalue())
        self.assertTrue("my_rom[1] = 3'h2;" in buffer.getvalue())
        self.assertTrue("my_rom[2] = 3'h3;" in buffer.getvalue())
        self.assertTrue("my_rom[3] = 3'h4;" in buffer.getvalue())

        buffer = io.StringIO()
        pyrtl.output_verilog_testbench(buffer, add_reset=False)
        # The testbench should not touch the RomBlock.
        self.assertTrue("my_rom" not in buffer.getvalue())


firrtl_output_concat_test = """\
circuit Example :
  module Example :
    input clock : Clock
    input reset : UInt<1>
    output o : UInt<13>
    wire tmp0 : UInt<13>
    wire tmp1 : UInt<7>
    wire tmp2 : UInt<13>
    node const_0_12 = UInt<4>(12)
    node const_1_3 = UInt<3>(3)
    node const_2_38 = UInt<6>(38)

    o <= tmp0
    tmp0 <= tmp2
    tmp1 <= cat(const_0_12, const_1_3)
    tmp2 <= cat(tmp1, const_2_38)
"""

firrtl_output_select_test = """\
circuit Example :
  module Example :
    input clock : Clock
    input reset : UInt<1>
    output b : UInt<6>
    wire tmp0 : UInt<6>
    wire tmp1 : UInt<1>
    wire tmp2 : UInt<1>
    wire tmp3 : UInt<1>
    wire tmp4 : UInt<1>
    wire tmp5 : UInt<1>
    wire tmp6 : UInt<1>
    wire tmp7 : UInt<6>
    wire tmp8 : UInt<2>
    wire tmp9 : UInt<3>
    wire tmp10 : UInt<4>
    wire tmp11 : UInt<5>
    wire tmp12 : UInt<6>
    node const_0_2893 = UInt<12>(2893)

    b <= tmp0
    tmp0 <= tmp7
    tmp1 <= bits(const_0_2893, 0, 0)
    tmp2 <= bits(const_0_2893, 2, 2)
    tmp3 <= bits(const_0_2893, 4, 4)
    tmp4 <= bits(const_0_2893, 6, 6)
    tmp5 <= bits(const_0_2893, 8, 8)
    tmp6 <= bits(const_0_2893, 10, 10)
    tmp7 <= tmp12
    tmp8 <= cat(tmp6, tmp5)
    tmp9 <= cat(tmp8, tmp4)
    tmp10 <= cat(tmp9, tmp3)
    tmp11 <= cat(tmp10, tmp2)
    tmp12 <= cat(tmp11, tmp1)
"""


class TestOutputFirrtl(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        pyrtl.wire._reset_wire_indexers()
        pyrtl.memory._reset_memory_indexer()

    def test_textual_consistency_concats(self):
        i = pyrtl.Const(0b1100)
        j = pyrtl.Const(0b011, bitwidth=3)
        k = pyrtl.Const(0b100110)
        o = pyrtl.Output(13, "o")
        o <<= pyrtl.concat(i, j, k)

        buffer = io.StringIO()
        pyrtl.output_to_firrtl(buffer)

        self.assertEqual(buffer.getvalue(), firrtl_output_concat_test)

    def test_textual_consistency_selects(self):
        a = pyrtl.Const(0b101101001101)
        b = pyrtl.Output(6, "b")
        b <<= a[::2]

        buffer = io.StringIO()
        pyrtl.output_to_firrtl(buffer)

        self.assertEqual(buffer.getvalue(), firrtl_output_select_test)


iscas85_bench_c432 = """\
# c432
# 36 inputs
# 7 outputs
# 40 inverters
# 120 gates ( 4 ANDs + 79 NANDs + 19 NORs + 18 XORs )

INPUT(1)
INPUT(4)
INPUT(8)
INPUT(11)
INPUT(14)
INPUT(17)
INPUT(21)
INPUT(24)
INPUT(27)
INPUT(30)
INPUT(34)
INPUT(37)
INPUT(40)
INPUT(43)
INPUT(47)
INPUT(50)
INPUT(53)
INPUT(56)
INPUT(60)
INPUT(63)
INPUT(66)
INPUT(69)
INPUT(73)
INPUT(76)
INPUT(79)
INPUT(82)
INPUT(86)
INPUT(89)
INPUT(92)
INPUT(95)
INPUT(99)
INPUT(102)
INPUT(105)
INPUT(108)
INPUT(112)
INPUT(115)

OUTPUT(223)
OUTPUT(329)
OUTPUT(370)
OUTPUT(421)
OUTPUT(430)
OUTPUT(431)
OUTPUT(432)

118 = NOT(1)
119 = NOT(4)
122 = NOT(11)
123 = NOT(17)
126 = NOT(24)
127 = NOT(30)
130 = NOT(37)
131 = NOT(43)
134 = NOT(50)
135 = NOT(56)
138 = NOT(63)
139 = NOT(69)
142 = NOT(76)
143 = NOT(82)
146 = NOT(89)
147 = NOT(95)
150 = NOT(102)
151 = NOT(108)
154 = NAND(118, 4)
157 = NOR(8, 119)
158 = NOR(14, 119)
159 = NAND(122, 17)
162 = NAND(126, 30)
165 = NAND(130, 43)
168 = NAND(134, 56)
171 = NAND(138, 69)
174 = NAND(142, 82)
177 = NAND(146, 95)
180 = NAND(150, 108)
183 = NOR(21, 123)
184 = NOR(27, 123)
185 = NOR(34, 127)
186 = NOR(40, 127)
187 = NOR(47, 131)
188 = NOR(53, 131)
189 = NOR(60, 135)
190 = NOR(66, 135)
191 = NOR(73, 139)
192 = NOR(79, 139)
193 = NOR(86, 143)
194 = NOR(92, 143)
195 = NOR(99, 147)
196 = NOR(105, 147)
197 = NOR(112, 151)
198 = NOR(115, 151)
199 = AND(154, 159, 162, 165, 168, 171, 174, 177, 180)
203 = NOT(199)
213 = NOT(199)
223 = NOT(199)
224 = XOR(203, 154)
227 = XOR(203, 159)
230 = XOR(203, 162)
233 = XOR(203, 165)
236 = XOR(203, 168)
239 = XOR(203, 171)
242 = NAND(1, 213)
243 = XOR(203, 174)
246 = NAND(213, 11)
247 = XOR(203, 177)
250 = NAND(213, 24)
251 = XOR(203, 180)
254 = NAND(213, 37)
255 = NAND(213, 50)
256 = NAND(213, 63)
257 = NAND(213, 76)
258 = NAND(213, 89)
259 = NAND(213, 102)
260 = NAND(224, 157)
263 = NAND(224, 158)
264 = NAND(227, 183)
267 = NAND(230, 185)
270 = NAND(233, 187)
273 = NAND(236, 189)
276 = NAND(239, 191)
279 = NAND(243, 193)
282 = NAND(247, 195)
285 = NAND(251, 197)
288 = NAND(227, 184)
289 = NAND(230, 186)
290 = NAND(233, 188)
291 = NAND(236, 190)
292 = NAND(239, 192)
293 = NAND(243, 194)
294 = NAND(247, 196)
295 = NAND(251, 198)
296 = AND(260, 264, 267, 270, 273, 276, 279, 282, 285)
300 = NOT(263)
301 = NOT(288)
302 = NOT(289)
303 = NOT(290)
304 = NOT(291)
305 = NOT(292)
306 = NOT(293)
307 = NOT(294)
308 = NOT(295)
309 = NOT(296)
319 = NOT(296)
329 = NOT(296)
330 = XOR(309, 260)
331 = XOR(309, 264)
332 = XOR(309, 267)
333 = XOR(309, 270)
334 = NAND(8, 319)
335 = XOR(309, 273)
336 = NAND(319, 21)
337 = XOR(309, 276)
338 = NAND(319, 34)
339 = XOR(309, 279)
340 = NAND(319, 47)
341 = XOR(309, 282)
342 = NAND(319, 60)
343 = XOR(309, 285)
344 = NAND(319, 73)
345 = NAND(319, 86)
346 = NAND(319, 99)
347 = NAND(319, 112)
348 = NAND(330, 300)
349 = NAND(331, 301)
350 = NAND(332, 302)
351 = NAND(333, 303)
352 = NAND(335, 304)
353 = NAND(337, 305)
354 = NAND(339, 306)
355 = NAND(341, 307)
356 = NAND(343, 308)
357 = AND(348, 349, 350, 351, 352, 353, 354, 355, 356)
360 = NOT(357)
370 = NOT(357)
371 = NAND(14, 360)
372 = NAND(360, 27)
373 = NAND(360, 40)
374 = NAND(360, 53)
375 = NAND(360, 66)
376 = NAND(360, 79)
377 = NAND(360, 92)
378 = NAND(360, 105)
379 = NAND(360, 115)
380 = NAND(4, 242, 334, 371)
381 = NAND(246, 336, 372, 17)
386 = NAND(250, 338, 373, 30)
393 = NAND(254, 340, 374, 43)
399 = NAND(255, 342, 375, 56)
404 = NAND(256, 344, 376, 69)
407 = NAND(257, 345, 377, 82)
411 = NAND(258, 346, 378, 95)
414 = NAND(259, 347, 379, 108)
415 = NOT(380)
416 = AND(381, 386, 393, 399, 404, 407, 411, 414)
417 = NOT(393)
418 = NOT(404)
419 = NOT(407)
420 = NOT(411)
421 = NOR(415, 416)
422 = NAND(386, 417)
425 = NAND(386, 393, 418, 399)
428 = NAND(399, 393, 419)
429 = NAND(386, 393, 407, 420)
430 = NAND(381, 386, 422, 399)
431 = NAND(381, 386, 425, 428)
432 = NAND(381, 422, 425, 429)
"""

iscas89_bench_s208 = """\
# s208
# 11 inputs
# 2 outputs
# 8 D-type flipflops
# 35 inverters
# 61 gates (17 ANDs + 19 NANDs + 4 ORs + 21 NORs)

INPUT(X)
INPUT(Clear)
INPUT(C_8)
INPUT(C_7)
INPUT(C_6)
INPUT(C_5)
INPUT(C_4)
INPUT(C_3)
INPUT(C_2)
INPUT(C_1)
INPUT(C_0)

OUTPUT(W)
OUTPUT(Z)

Y_4 = DFF(II3)
Y_3 = DFF(II4)
Y_2 = DFF(II5)
Y_1 = DFF(II6)
Y_8 = DFF(II155)
Y_7 = DFF(II156)
Y_6 = DFF(II157)
Y_5 = DFF(II158)

II6 = NOT(II104)
II50 = NOT(II92)
II40 = NOT(X)
II41 = NOT(Y_4)
II42 = NOT(Y_3)
II43 = NOT(Y_2)
II44 = NOT(Y_1)
II3 = NOT(II46)
II4 = NOT(II47)
II158 = NOT(II256)
II202 = NOT(II244)
II192 = NOT(II1_1)
II193 = NOT(Y_8)
II194 = NOT(Y_7)
II195 = NOT(Y_6)
II196 = NOT(Y_5)
II155 = NOT(II198)
II156 = NOT(II199)
II307_1 = NOT(II341)
II318 = NOT(II341)
II309 = NOT(Y_1)
II310 = NOT(Y_2)
II311 = NOT(Y_3)
P_1 = NOT(II314)
P_3 = NOT(II316)
P_4 = NOT(II317)
II368 = NOT(Y_4)
II369 = NOT(Y_5)
II370 = NOT(Y_6)
II371 = NOT(Y_7)
II372 = NOT(Y_8)
P_6 = NOT(II374)
Z = NOT(II446)
II487 = NOT(II488)
II489 = NOT(II490)

II127_1 = AND(II41, Y_3, II109)
II127_2 = AND(II96, II113, Y_4)
II131_1 = AND(II113, II92, Y_3)
II131_2 = AND(II42, II109)
II279_1 = AND(II193, Y_7, II261)
II279_2 = AND(II248, II265, Y_8)
II283_1 = AND(II265, II244, Y_7)
II283_2 = AND(II194, II261)
II497_1 = AND(P_8, C_8)
II500_1 = AND(P_5, C_5)
II500_2 = AND(P_2, C_2)
II504_1 = AND(P_3, C_3)
II504_2 = AND(P_4, C_4)
II508_1 = AND(X, C_0)
II508_2 = AND(P_1, C_1)
II512_1 = AND(P_6, C_6)
II512_2 = AND(P_7, C_7)

II135_1 = OR(II43, II104)
II135_2 = OR(Y_2, II100)
II287_1 = OR(II195, II256)
II287_2 = OR(Y_6, II252)

II5 = NAND(II135_1, II135_2)
II92 = NAND(Y_2, Y_1)
II96 = NAND(Y_3, II50)
II100 = NAND(Y_1, II113)
II104 = NAND(II44, II113)
II157 = NAND(II287_1, II287_2)
II244 = NAND(Y_6, Y_5)
II248 = NAND(Y_7, II202)
II252 = NAND(Y_5, II265)
II256 = NAND(II196, II265)
II314 = NAND(X, Y_1)
II316 = NAND(Y_3, II347)
II317 = NAND(Y_4, II318)
II341 = NAND(II311, II347)
II350 = NAND(X, II309)
II374 = NAND(Y_6, II406)
II378 = NAND(II406, II370)
II409 = NAND(II368, II307_1)
II495 = NAND(II484, II494)

II1_1 = NOR(II41, II96)
II46 = NOR(II127_1, II127_2)
II47 = NOR(II131_1, II131_2)
II109 = NOR(II43, II100)
II113 = NOR(Clear, II40)
W = NOR(II193, II248)
II198 = NOR(II279_1, II279_2)
II199 = NOR(II283_1, II283_2)
II261 = NOR(II195, II252)
II265 = NOR(Clear, II192)
P_2 = NOR(II310, II350)
II347 = NOR(Y_2, II350)
P_5 = NOR(II369, II409)
P_7 = NOR(II371, II378)
P_8 = NOR(Y_7, II378, II372)
II406 = NOR(II409, Y_5)
II446 = NOR(II495, II487, II497_1)
II484 = NOR(II500_1, II500_2)
II488 = NOR(II504_1, II504_2)
II490 = NOR(II508_1, II508_2)
II494 = NOR(II512_1, II512_2, II489)
"""

iscas89_bench_s27 = """\
# 4 inputs
# 1 outputs
# 3 D-type flipflops
# 2 inverters
# 8 gates (1 ANDs + 1 NANDs + 2 ORs + 4 NORs)

INPUT(G0)
INPUT(G1)
INPUT(G2)
INPUT(G3)

OUTPUT(G17)

G5 = DFF(G10)
G6 = DFF(G11)
G7 = DFF(G13)

G14 = NOT(G0)
G17 = NOT(G11)

G8 = AND(G14, G6)

G15 = OR(G12, G8)
G16 = OR(G3, G8)

G9 = NAND(G16, G15)

G10 = NOR(G14, G11)
G11 = NOR(G5, G9)
G12 = NOR(G1, G7)
G13 = NOR(G2, G12)
"""

example_bench_with_io_same_name = """\
INPUT(G1)
INPUT(G2)
INPUT(G3)

OUTPUT(G3)
OUTPUT(G4)

G4 = OR(G1, G2)
"""


class TestInputISCASBench(unittest.TestCase):
    # NOTE: number of inverters = number of original inverters + number of NORs, since
    # we currently convert NORs to inverters and ORs. Similarly for number of total ORs.

    def setUp(self):
        pyrtl.reset_working_block()
        pyrtl.wire._reset_wire_indexers()
        pyrtl.memory._reset_memory_indexer()

    def check_io(self, cls, names):
        block = pyrtl.working_block()
        for name in names:
            wire = block.get_wirevector_by_name(name)
            self.assertIsNotNone(wire)
            self.assertIsInstance(wire, cls)
            self.assertEqual(len(wire), 1)
        self.assertEqual(len(block.wirevector_subset(cls)), len(names))

    def check_gate_amounts(self, nots, ands, ors, nands, xors, dffs):
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic_subset(op="~")), nots)
        self.assertEqual(len(block.logic_subset(op="&")), ands)
        self.assertEqual(len(block.logic_subset(op="|")), ors)
        self.assertEqual(len(block.logic_subset(op="n")), nands)
        self.assertEqual(len(block.logic_subset(op="^")), xors)
        self.assertEqual(len(block.wirevector_subset(pyrtl.Register)), dffs)

    def test_combinational_bench(self):
        pyrtl.input_from_iscas_bench(iscas85_bench_c432)

        input_names = [
            "1",
            "4",
            "8",
            "11",
            "14",
            "17",
            "21",
            "24",
            "27",
            "30",
            "34",
            "37",
            "40",
            "43",
            "47",
            "50",
            "53",
            "56",
            "60",
            "63",
            "66",
            "69",
            "73",
            "76",
            "79",
            "82",
            "86",
            "89",
            "92",
            "95",
            "99",
            "102",
            "105",
            "108",
            "112",
            "115",
        ]
        self.check_io(pyrtl.Input, input_names)

        output_names = ["223", "329", "370", "421", "430", "431", "432"]
        self.check_io(pyrtl.Output, output_names)

        self.check_gate_amounts(59, 4, 19, 79, 18, 0)

    def test_sequential_bench(self):
        pyrtl.input_from_iscas_bench(iscas89_bench_s208)

        input_names = [
            "X",
            "Clear",
            "C_8",
            "C_7",
            "C_6",
            "C_5",
            "C_4",
            "C_3",
            "C_2",
            "C_1",
            "C_0",
        ]
        self.check_io(pyrtl.Input, input_names)

        output_names = ["W", "Z"]
        self.check_io(pyrtl.Output, output_names)

        self.check_gate_amounts(56, 17, 25, 19, 0, 8)

    def test_simulation_bench(self):
        pyrtl.input_from_iscas_bench(iscas89_bench_s27)
        trace = pyrtl.SimulationTrace(
            {pyrtl.working_block().get_wirevector_by_name("G17")}
        )
        sim = pyrtl.Simulation(trace)
        sim.step_multiple(
            {
                "G0": "01101001",
                "G1": "11010100",
                "G2": "00011001",
                "G3": "11001100",
            }
        )
        correct_output = "G17 11110001\n"
        output = io.StringIO()
        trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_output)

    def test_bench_with_same_io_name(self):
        output = io.StringIO()
        sys.stdout = output
        pyrtl.input_from_iscas_bench(example_bench_with_io_same_name)
        sys.stdout = sys.__stdout__

        self.assertEqual(
            output.getvalue(),
            "Found input and output wires with the same name. Output 'G3' has now been "
            "renamed to 'tmp3'.\n",
        )
        pyrtl.working_block().sanity_check()

        self.check_io(pyrtl.Input, ["G1", "G2", "G3"])
        self.check_io(pyrtl.Output, ["tmp3", "G4"])


if __name__ == "__main__":
    unittest.main()



import unittest

import pyrtl


class NetWireNumTestCases(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def assert_num_net(self, num, block=None):
        block = pyrtl.working_block(block)
        self.assertEqual(len(block.logic), num)

    def assert_num_wires(self, num, block=None):
        block = pyrtl.working_block(block)
        self.assertEqual(len(block.wirevector_set), num)

    def num_net_of_type(self, netOp, num, block=None):
        block = pyrtl.working_block(block)
        self.assertEqual(len([net for net in block.logic if net.op == netOp]), num)

    def num_wire_of_type(self, wiretype, num, block=None):
        block = pyrtl.working_block(block)
        self.assertEqual(len(block.wirevector_subset(wiretype)), num)


class WireMemoryNameTestCases(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def name_wires(self, names, block=None):
        block = pyrtl.working_block()
        names = set(names.split(" "))
        for n in names:
            self.assertIn(n, block.wirevector_by_name)

    def name_memories(self, names, block=None):
        block = pyrtl.working_block()
        names = set(names.split(" "))
        for n in names:
            self.assertIn(n, block.memblock_by_name)


def insert_random_inversions(rate=0.5):
    """
    an example transform that can be used for testing
    """

    import random

    def randomly_replace(wire):
        if random.random() < rate:
            new_src = pyrtl.clone_wire(wire, pyrtl.wire.next_tempvar_name())
            new_dst = pyrtl.clone_wire(wire, pyrtl.wire.next_tempvar_name())
            new_dst <<= ~new_src
            return new_src, new_dst
        return wire, wire

    pyrtl.wire_transform(randomly_replace)


class TestWireTransform(NetWireNumTestCases):
    def test_randomly_replace(self):
        a, b = pyrtl.WireVector(3), pyrtl.WireVector(3)
        o = a & b
        insert_random_inversions(1)
        block = pyrtl.working_block()
        self.num_net_of_type("~", 3, block)
        self.num_net_of_type("&", 1, block)

        new_and_net = block.logic_subset("&").pop()
        for arg in new_and_net.args:
            self.assertIsNot(arg, a)
            self.assertIsNot(arg, b)
        self.assertIsNot(new_and_net.dests[0], o)

    def test_replace_input(self):
        def f(wire):
            if wire.name == "a":
                w = pyrtl.clone_wire(wire, "w2")
            else:
                w = pyrtl.clone_wire(wire, "w3")
            return wire, w

        a, b = pyrtl.input_list("a/1 b/1")
        w1 = a & b
        o = pyrtl.Output(1, "o")
        o <<= w1

        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertEqual(src_nets[w1], pyrtl.LogicNet("&", None, (a, b), (w1,)))
        self.assertIn(a, dst_nets)
        self.assertIn(b, dst_nets)

        pyrtl.wire_transform(f, select_types=pyrtl.Input, exclude_types=())

        w2 = pyrtl.working_block().get_wirevector_by_name("w2")
        w3 = pyrtl.working_block().get_wirevector_by_name("w3")
        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertEqual(src_nets[w1], pyrtl.LogicNet("&", None, (w2, w3), (w1,)))
        self.assertNotIn(a, dst_nets)
        self.assertNotIn(b, dst_nets)

    def test_replace_output(self):
        def f(wire):
            w = pyrtl.clone_wire(wire, "w2")
            return w, wire

        a, b = pyrtl.input_list("a/1 b/1")
        w1 = a & b
        o = pyrtl.Output(1, "o")
        o <<= w1

        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertEqual(dst_nets[w1], [pyrtl.LogicNet("w", None, (w1,), (o,))])
        self.assertIn(o, src_nets)

        pyrtl.wire_transform(f, select_types=pyrtl.Output, exclude_types=())

        w2 = pyrtl.working_block().get_wirevector_by_name("w2")
        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertEqual(dst_nets[w1], [pyrtl.LogicNet("w", None, (w1,), (w2,))])
        self.assertNotIn(o, src_nets)


class TestCopyBlock(NetWireNumTestCases, WireMemoryNameTestCases):
    def num_memories(self, mems_expected, block):
        memories = set()
        for net in block.logic_subset("m@"):
            memories.add(net.op_param[1])  # location of the memories object
        self.assertEqual(mems_expected, len(memories))

    def test_blank(self):
        block = pyrtl.copy_block()
        self.assert_num_net(0, block)
        self.assert_num_wires(0, block)

    def test_block(self):
        a = pyrtl.Const(23, name="a")
        b = pyrtl.Input(5, name="b")
        o = pyrtl.Output(5, name="o")
        o <<= ~a & b

        old_block = pyrtl.working_block()
        old_block.sanity_check()
        self.assert_num_wires(5, old_block)
        self.assert_num_net(3, old_block)

        self.name_wires("a b o", old_block)

        new_block = pyrtl.copy_block()
        new_block.sanity_check()
        self.assert_num_wires(5, new_block)
        self.assert_num_net(3, new_block)

        self.name_wires("a b o", new_block)

    def test_copy_mem(self):
        ins = [pyrtl.Input(5) for i in range(4)]
        out = pyrtl.Output(5)

        mem1 = pyrtl.MemBlock(5, 5, name="mem1")
        mem2 = pyrtl.MemBlock(5, 5, name="mem2")

        mem1_o1 = mem1[ins[0]]
        mem1[ins[1]] <<= ins[2]
        mem2_o2 = mem2[ins[3]]
        out <<= mem1_o1 & mem2_o2

        old_block = pyrtl.working_block()
        old_block.sanity_check()
        self.num_net_of_type("m", 2, old_block)
        self.num_net_of_type("@", 1, old_block)
        self.num_net_of_type("&", 1, old_block)
        self.num_memories(2, old_block)

        self.name_memories("mem1 mem2", old_block)

        new_block = pyrtl.copy_block()
        self.num_net_of_type("m", 2, new_block)
        self.num_net_of_type("@", 1, new_block)
        self.num_net_of_type("&", 1, new_block)
        self.num_memories(2, new_block)

        self.name_memories("mem1 mem2", new_block)


class TestFastWireReplace(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_replace_multiple_wires(self):
        j, n = pyrtl.Input(8), pyrtl.Output(8)
        o, h = pyrtl.WireVector(), pyrtl.WireVector()
        x, y = pyrtl.WireVector(8), pyrtl.WireVector(8)

        o <<= j
        h <<= o
        n <<= h
        block = pyrtl.working_block()
        src_nets, dst_nets = block.net_connections()
        pyrtl.replace_wire_fast(o, x, x, src_nets, dst_nets)
        pyrtl.replace_wire_fast(h, y, y, src_nets, dst_nets)
        for old_wire in (o, h):
            self.assertNotIn(old_wire, src_nets)
            self.assertNotIn(old_wire, dst_nets)
            self.assertNotIn(old_wire, block.wirevector_set)
        block.sanity_check()

    def test_replace_multiple_wires_2(self):
        j, n = pyrtl.Input(8), pyrtl.Output(8)
        o = pyrtl.WireVector()
        x, y, z = pyrtl.WireVector(8), pyrtl.WireVector(8), pyrtl.WireVector(8)

        o <<= j
        p = ~j
        h = o & p
        n <<= h
        block = pyrtl.working_block()
        src_nets, dst_nets = block.net_connections()
        pyrtl.replace_wire_fast(o, x, x, src_nets, dst_nets)
        pyrtl.replace_wire_fast(p, z, z, src_nets, dst_nets)
        pyrtl.replace_wire_fast(h, y, y, src_nets, dst_nets)
        for old_wire in (o, h, p):
            self.assertNotIn(old_wire, src_nets)
            self.assertNotIn(old_wire, dst_nets)
            self.assertNotIn(old_wire, block.wirevector_set)
        block.sanity_check()

    def test_wire_used_in_multiple_places(self):
        j, k = pyrtl.Input(8), pyrtl.Input(8)
        n, o = pyrtl.Output(8), pyrtl.Output(8)
        x = pyrtl.WireVector(8)

        r = j & k
        n <<= j | r
        o <<= r ^ k

        block = pyrtl.working_block()
        src_nets, dst_nets = block.net_connections()
        pyrtl.replace_wire_fast(r, x, x, src_nets, dst_nets)

        for old_wire in (r,):
            self.assertNotIn(old_wire, src_nets)
            self.assertNotIn(old_wire, dst_nets)
            self.assertNotIn(old_wire, block.wirevector_set)
        block.sanity_check()

    def test_replace_only_src_wire(self):
        a, b, c, d = pyrtl.input_list("a/1 b/1 c/1 d/1")
        w1 = a & b
        w1.name = "w1"
        w2 = c | d
        w2.name = "w2"
        w3 = w1 ^ w2
        w3.name = "w3"
        o = pyrtl.Output(1, "o")
        o <<= w3

        w4 = pyrtl.WireVector(1, "w4")
        src_nets, dst_nets = pyrtl.working_block().net_connections()

        w1_src_net = src_nets[w1]
        w1_dst_net = dst_nets[w1][0]
        self.assertEqual(w1_src_net.args, (a, b))
        self.assertEqual(w1_src_net.dests, (w1,))
        self.assertEqual(w1_dst_net.args, (w1, w2))
        self.assertEqual(w1_dst_net.dests, (w3,))
        self.assertNotIn(w4, src_nets)

        pyrtl.replace_wire_fast(w1, w4, w1, src_nets, dst_nets)

        self.assertNotIn(w1, src_nets)  # The maps have been updated...
        self.assertEqual(dst_nets[w1], [w1_dst_net])
        w4_src_net = src_nets[
            w4
        ]  # ...but the net can't be, so new updated versions were created
        self.assertEqual(w4_src_net.args, w1_src_net.args)
        self.assertEqual(w4_src_net.dests, (w4,))

    def test_replace_only_dst_wire(self):
        a, b, c, d = pyrtl.input_list("a/1 b/1 c/1 d/1")
        w1 = a & b
        w1.name = "w1"
        w2 = c | d
        w2.name = "w2"
        w3 = w1 ^ w2
        w3.name = "w3"
        o = pyrtl.Output(1, "o")
        o <<= w3

        w4 = pyrtl.WireVector(1, "w4")
        src_nets, dst_nets = pyrtl.working_block().net_connections()

        w1_src_net = src_nets[w1]
        w1_dst_net = dst_nets[w1][0]
        self.assertEqual(w1_src_net.args, (a, b))
        self.assertEqual(w1_src_net.dests, (w1,))
        self.assertEqual(w1_dst_net.args, (w1, w2))
        self.assertEqual(w1_dst_net.dests, (w3,))
        self.assertNotIn(w4, src_nets)

        pyrtl.replace_wire_fast(w1, w1, w4, src_nets, dst_nets)

        self.assertNotIn(w1, dst_nets)  # The maps have been updated...
        self.assertEqual(src_nets[w1], w1_src_net)
        w4_dst_net = dst_nets[w4][
            0
        ]  # ...but the net can't be, so new versions were created
        self.assertEqual(w4_dst_net.args, (w4, w2))
        self.assertEqual(w4_dst_net.dests, w1_dst_net.dests)


class TestCloning(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_same_type(self):
        for ix, cls in enumerate(
            [pyrtl.WireVector, pyrtl.Register, pyrtl.Input, pyrtl.Output]
        ):
            w1 = cls(4, f"w{ix}")
            w2 = pyrtl.clone_wire(w1, f"y{ix}")
            self.assertIsInstance(w2, cls)
            self.assertEqual(w1.bitwidth, w2.bitwidth)

    def test_clone_wire_no_name_same_block(self):
        a = pyrtl.WireVector(1, "a")
        with self.assertRaises(pyrtl.PyrtlError) as error:
            pyrtl.clone_wire(a)
        self.assertEqual(
            str(error.exception),
            "Must provide a name for the newly cloned wire when cloning within the "
            "same block.",
        )

    def test_clone_wire_same_name_same_block(self):
        a = pyrtl.WireVector(1, "a")
        with self.assertRaises(pyrtl.PyrtlError) as error:
            pyrtl.clone_wire(a, "a")
        self.assertEqual(
            str(error.exception),
            "Cannot give a newly cloned wire the same name as an existing wire.",
        )

    def test_clone_wire_different_name_same_block(self):
        a = pyrtl.WireVector(1, "a")
        self.assertEqual(a.name, "a")
        self.assertEqual(pyrtl.working_block().wirevector_set, {a})
        self.assertIs(pyrtl.working_block().wirevector_by_name["a"], a)

        w = pyrtl.clone_wire(a, name="w")
        self.assertEqual(w.name, "w")
        self.assertEqual(a.name, "a")
        self.assertIs(pyrtl.working_block().wirevector_by_name["w"], w)
        self.assertIs(pyrtl.working_block().wirevector_by_name["a"], a)
        self.assertEqual(pyrtl.working_block().wirevector_set, {a, w})

        pyrtl.working_block().remove_wirevector(a)
        self.assertEqual(pyrtl.working_block().wirevector_set, {w})

    def test_clone_wire_no_or_same_name_different_block(self):
        for clone_name in (None, "a"):
            a = pyrtl.WireVector(1, "a")
            b = pyrtl.Block()
            with pyrtl.set_working_block(b):
                w = pyrtl.clone_wire(a, name=clone_name)

            self.assertEqual(a.name, "a")
            self.assertIs(pyrtl.working_block().wirevector_by_name["a"], a)
            self.assertEqual(pyrtl.working_block().wirevector_set, {a})

            self.assertEqual(w.name, "a")
            self.assertIs(b.wirevector_by_name["a"], w)
            self.assertEqual(b.wirevector_set, {w})
            pyrtl.reset_working_block()

    def test_clone_wire_different_name_different_block(self):
        a = pyrtl.WireVector(1, "a")
        b = pyrtl.Block()
        with pyrtl.set_working_block(b):
            w = pyrtl.clone_wire(a, "w")
        self.assertEqual(a.name, "a")
        self.assertEqual(w.name, "w")
        self.assertIs(pyrtl.working_block().wirevector_by_name["a"], a)
        self.assertEqual(pyrtl.working_block().wirevector_set, {a})
        self.assertIs(b.wirevector_by_name["w"], w)
        self.assertEqual(b.wirevector_set, {w})


# this code needs mocking from python 3's unittests to work
"""
@mock.patch('transform_examples.pyrtl.probe')
def test_probe(self, probe):
    # Note to readers, this is a rather contrived test
    # If you want to know how to probe a single wirevector, look at the
    # probe function in pyrtl core
    in_wire, in_wire2 = pyrtl.Input(3), pyrtl.Input(3)
    out_wire = pyrtl.Output()
    test_wire = ~(in_wire & in_wire2)
    out_wire <<= test_wire

    def probe_cond(wire):
        return wire is test_wire

    transform_examples.probe_wire_if(probe_cond)
    probe.assert_called_once_with(test_wire)
"""

if __name__ == "__main__":
    unittest.main()



import io
import unittest

import pyrtl


class TestAreaEstimate(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_area_est_unchanged(self):
        a = pyrtl.Const(2, 8)
        b = pyrtl.Const(85, 8)
        zero = pyrtl.Const(0, 1)
        reg = pyrtl.Register(8)
        mem = pyrtl.MemBlock(8, 8)
        out = pyrtl.Output(8)
        (
            nota,
            aLSB,
            athenb,
            aORb,
            aANDb,
            aNANDb,
            aXORb,
            aequalsb,
            altb,
            agtb,
            aselectb,
            aplusb,
            bminusa,
            atimesb,
            memread,
        ) = [pyrtl.Output() for i in range(15)]
        out <<= zero
        nota <<= ~a
        aLSB <<= a[0]
        athenb <<= pyrtl.concat(a, b)
        aORb <<= a | b
        aANDb <<= a & b
        aNANDb <<= a.nand(b)
        aXORb <<= a ^ b
        aequalsb <<= a == b
        altb <<= a < b
        agtb <<= a > b
        aselectb <<= pyrtl.select(zero, a, b)
        reg.next <<= a
        aplusb <<= a + b
        bminusa <<= a - b
        atimesb <<= a * b
        memread <<= mem[0]
        mem[1] <<= a
        self.assertEqual(pyrtl.area_estimation(), (0.00734386752, 0.01879779717361501))

    def test_area_est_unchanged_with_rom(self):
        a = pyrtl.Const(2, 8)
        b = pyrtl.Const(85, 8)
        zero = pyrtl.Const(0, 1)
        reg = pyrtl.Register(8)
        mem = pyrtl.RomBlock(8, 8, romdata=list(range(0, 256)))
        out = pyrtl.Output(8)
        (
            nota,
            aLSB,
            athenb,
            aORb,
            aANDb,
            aNANDb,
            aXORb,
            aequalsb,
            altb,
            agtb,
            aselectb,
            aplusb,
            bminusa,
            atimesb,
            memread,
        ) = [pyrtl.Output() for i in range(15)]
        out <<= zero
        nota <<= ~a
        aLSB <<= a[0]
        athenb <<= pyrtl.concat(a, b)
        aORb <<= a | b
        aANDb <<= a & b
        aNANDb <<= a.nand(b)
        aXORb <<= a ^ b
        aequalsb <<= a == b
        altb <<= a < b
        agtb <<= a > b
        aselectb <<= pyrtl.select(zero, a, b)
        reg.next <<= a
        aplusb <<= a + b
        bminusa <<= a - b
        atimesb <<= a * b
        memread <<= mem[reg]
        self.assertEqual(pyrtl.area_estimation(), (0.00734386752, 0.001879779717361501))


class TestTimingEstimate(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_time_est_unchanged(self):
        a = pyrtl.Const(2, 8)
        b = pyrtl.Const(85, 8)
        zero = pyrtl.Const(0, 1)
        reg = pyrtl.Register(8)
        mem = pyrtl.MemBlock(8, 8)
        out = pyrtl.Output(8)
        (
            nota,
            aLSB,
            athenb,
            aORb,
            aANDb,
            aNANDb,
            aXORb,
            aequalsb,
            altb,
            agtb,
            aselectb,
            aplusb,
            bminusa,
            atimesb,
            memread,
        ) = [pyrtl.Output() for i in range(15)]
        out <<= zero
        nota <<= ~a
        aLSB <<= a[0]
        athenb <<= pyrtl.concat(a, b)
        aORb <<= a | b
        aANDb <<= a & b
        aNANDb <<= a.nand(b)
        aXORb <<= a ^ b
        aequalsb <<= a == b
        altb <<= a < b
        agtb <<= a > b
        aselectb <<= pyrtl.select(zero, a, b)
        reg.next <<= a
        aplusb <<= a + b
        bminusa <<= a - b
        atimesb <<= a * b
        memread <<= mem[0]
        mem[1] <<= a
        timing = pyrtl.TimingAnalysis()
        self.assertEqual(timing.max_freq(), 610.2770657878676)
        self.assertEqual(timing.max_length(), 1255.6000000000001)


class TestYosysInterface(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()


paths_print_output = """\
From i
  To o
    Path 0
      tmp5/3W <-- - -- i/2I, tmp4/2W
      tmp6/3W <-- | -- tmp2/3W, tmp5/3W
      o/3O <-- w -- tmp6/3W
    Path 1
      tmp1/3W <-- c -- tmp0/1W, i/2I
      tmp2/3W <-- & -- tmp1/3W, j/3I
      tmp6/3W <-- | -- tmp2/3W, tmp5/3W
      o/3O <-- w -- tmp6/3W
  To p
    Path 0
      tmp8/4W <-- c -- tmp7/2W, i/2I
      tmp9/5W <-- - -- k/4I, tmp8/4W
      p/5O <-- w -- tmp9/5W
From j
  To o
    Path 0
      tmp2/3W <-- & -- tmp1/3W, j/3I
      tmp6/3W <-- | -- tmp2/3W, tmp5/3W
      o/3O <-- w -- tmp6/3W
  To p
    (No paths)
From k
  To o
    (No paths)
  To p
    Path 0
      tmp9/5W <-- - -- k/4I, tmp8/4W
      p/5O <-- w -- tmp9/5W
"""


class TestPaths(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        # To compare textual consistency, need to make sure we're starting at the same
        # index for all automatically created names.
        pyrtl.wire._reset_wire_indexers()
        pyrtl.memory._reset_memory_indexer()

    def test_one_path_to_one_output(self):
        a = pyrtl.Input(4, "a")
        o = pyrtl.Output(name="o")
        o <<= a * 2

        paths = pyrtl.paths(a, o)
        paths_a_to_o = paths[a][o]
        self.assertEqual(len(paths_a_to_o), 1)
        path_a_to_o = paths_a_to_o[0]
        self.assertEqual(len(path_a_to_o), 2)
        self.assertEqual(path_a_to_o[0].op, "*")
        self.assertEqual(path_a_to_o[1].op, "w")

    def test_two_paths_to_one_output(self):
        a = pyrtl.Input(4, "a")
        w1 = a * 2
        w2 = a + 1
        o = pyrtl.Output(name="o")
        o <<= w1 - w2

        paths = pyrtl.paths(a, o)
        self.assertEqual(len(paths[a][o]), 2)
        for path in paths[a][o]:
            if path[0].op == "*":
                self.assertEqual(len(path), 3)
            else:
                # Has an extra 'c'
                self.assertEqual(path[0].op, "+")
                self.assertEqual(len(path), 4)

    def test_two_paths_to_two_outputs(self):
        a = pyrtl.Input(4, "a")
        o1, o2 = pyrtl.output_list("o1/5 o1/8")
        o1 <<= a + 1
        o2 <<= a * 2

        paths = pyrtl.paths(a)
        paths_from_a = paths[a]
        self.assertEqual(len(paths_from_a), 2)
        path_to_o1 = paths_from_a[o1]
        self.assertEqual(len(path_to_o1), 1)
        path_to_o2 = paths_from_a[o2]
        self.assertEqual(len(path_to_o2), 1)

    def test_subset_of_all_paths(self):
        i, j, k = pyrtl.input_list("i/2 j/3 k/4")
        o, p = pyrtl.Output(), pyrtl.Output()
        o <<= i & j
        p <<= k - i

        # Make sure passing in both set and list works
        paths = pyrtl.paths([i, k], {o})
        paths_from_i = paths[i]
        self.assertNotIn(p, paths_from_i)  # Because p was not provided as target output
        self.assertEqual(len(paths_from_i[o]), 1)  # One path from i to o
        self.assertEqual(paths_from_i[o][0][0].op, "c")
        self.assertEqual(paths_from_i[o][0][1].op, "&")
        self.assertEqual(paths_from_i[o][0][2].op, "w")

        paths_from_k = paths[k]
        self.assertNotIn(p, paths_from_k)  # Because p was not provided as target output
        self.assertEqual(len(paths_from_k[o]), 0)  # 0 paths from k to o

    def test_paths_empty_src_and_dst_equal_with_no_other_logic(self):
        i = pyrtl.Input(4, "i")
        paths = pyrtl.paths(i, i)
        self.assertEqual(len(paths[i][i]), 0)

    def test_paths_with_loop(self):
        r = pyrtl.Register(1, "r")
        r.next <<= r & ~r
        paths = pyrtl.paths(r, r)
        self.assertEqual(len(paths[r][r]), 2)
        p1, p2 = sorted(paths[r][r], key=lambda p: len(p), reverse=True)
        self.assertEqual(len(p1), 3)
        self.assertEqual(p1[0].op, "~")
        self.assertEqual(p1[1].op, "&")
        self.assertEqual(p1[2].op, "r")
        self.assertEqual(len(p2), 2)
        self.assertEqual(p2[0].op, "&")
        self.assertEqual(p2[1].op, "r")

    def test_paths_loop_and_input(self):
        i = pyrtl.Input(1, "i")
        o = pyrtl.Output(1, "o")
        r = pyrtl.Register(1, "r")
        r.next <<= i & r
        o <<= r
        paths = pyrtl.paths(r, o)
        self.assertEqual(len(paths[r][o]), 1)

    def test_paths_loop_get_arbitrary_inner_wires(self):
        w = pyrtl.WireVector(1, "w")
        y = w & pyrtl.Const(1)
        w <<= ~y
        paths = pyrtl.paths(w, y)
        self.assertEqual(len(paths[w][y]), 1)
        self.assertEqual(paths[w][y][0][0].op, "&")

    def test_paths_no_path_exists(self):
        i = pyrtl.Input(1, "i")
        o = pyrtl.Output(1, "o")
        o <<= ~i

        w = pyrtl.WireVector(1, "w")
        y = w & pyrtl.Const(1)
        w <<= ~y

        paths = pyrtl.paths(w, o)
        self.assertEqual(len(paths[w][o]), 0)

    def test_paths_with_memory(self):
        i = pyrtl.Input(4, "i")
        o = pyrtl.Output(8, "o")
        mem = pyrtl.MemBlock(8, 32, "mem")
        waddr = pyrtl.Input(32, "waddr")
        raddr = pyrtl.Input(32, "raddr")
        data = mem[raddr]
        mem[waddr] <<= (i + ~data).truncate(8)
        o <<= data

        paths = pyrtl.paths(i, o)
        path = paths[i][o][0]
        self.assertEqual(path[0].op, "c")
        self.assertEqual(path[1].op, "+")
        self.assertEqual(path[2].op, "s")
        self.assertEqual(path[3].op, "@")
        self.assertEqual(path[4].op, "m")
        self.assertEqual(path[5].op, "w")

        # TODO Once issue with _MemIndexed lookups is resolved, these should be `data`
        # instead of `data.wire`.
        paths = pyrtl.paths(data.wire, data.wire)
        path = paths[data.wire][data.wire][0]
        self.assertEqual(path[0].op, "~")
        self.assertEqual(path[1].op, "+")
        self.assertEqual(path[2].op, "s")
        self.assertEqual(path[3].op, "@")
        self.assertEqual(path[4].op, "m")

    def test_all_paths(self):
        a, b, c = pyrtl.input_list("a/2 b/4 c/1")
        o, p = pyrtl.output_list("o/4 p/2")
        o <<= a + (b ^ (b + 1))
        p <<= c * 2 - a

        paths = pyrtl.paths()

        # We have entries for every input, output pair
        for start in (a, b, c):
            self.assertEqual(len(paths[start]), 2)
            self.assertTrue([w.name for w in paths[start]], [o.name, p.name])

        paths_a_to_o = paths[a][o]
        self.assertEqual(len(paths_a_to_o), 1)
        path_a_to_o = paths_a_to_o[0]
        self.assertEqual(path_a_to_o[0].op, "c")
        self.assertEqual(path_a_to_o[1].op, "+")
        self.assertEqual(path_a_to_o[2].op, "s")

        paths_a_to_p = paths[a][p]
        self.assertEqual(len(paths_a_to_p), 1)
        path_a_to_p = paths_a_to_p[0]
        self.assertEqual(path_a_to_p[0].op, "c")
        self.assertEqual(path_a_to_p[1].op, "-")
        self.assertEqual(path_a_to_p[2].op, "s")

        paths_b_to_o = paths[b][o]
        self.assertEqual(len(paths_b_to_o), 2)
        paths_b_to_p = paths[b][p]
        self.assertEqual(len(paths_b_to_p), 0)

        paths_c_to_o = paths[c][o]
        self.assertEqual(len(paths_c_to_o), 0)
        paths_c_to_p = paths[c][p]
        self.assertEqual(len(paths_c_to_p), 1)

    def test_pretty_print(self):
        i, j, k = pyrtl.input_list("i/2 j/3 k/4")
        o, p = pyrtl.Output(name="o"), pyrtl.Output(name="p")
        o <<= (i & j) | (i - 1)
        p <<= k - i

        paths = pyrtl.paths()
        output = io.StringIO()
        paths.print(file=output)
        self.assertEqual(output.getvalue(), paths_print_output)


class TestDistance(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_simple_distance(self):
        for b in (None, pyrtl.working_block()):
            a = pyrtl.Input(4, "a")
            o = pyrtl.Output(name="o")
            o <<= a * 2

            distances = pyrtl.distance(a, o, lambda _: 1, b)
            self.assertEqual(len(distances), 1)
            self.assertEqual(next(iter(distances.values())), 2)

    def test_several_distances(self):
        a = pyrtl.Input(4, "a")
        o = pyrtl.Output(name="o")
        w1 = a * 2
        w2 = w1 & 0b1101 + a
        o <<= ~w2

        distances = pyrtl.distance(a, o, lambda _: 1)
        self.assertEqual(len(distances), 2)
        for path, distance in distances.items():
            self.assertEqual(distance, len(path))

    def test_special_cost(self):
        a = pyrtl.Input(4, "a")
        o = pyrtl.Output(name="o")
        w1 = a * 2
        w2 = w1 - 4
        o <<= ~w2

        def cost(net):
            if net.op in "wcs":
                return 0
            if net.op in "*+-":
                return 2
            return 1

        distances = pyrtl.distance(a, o, cost)
        self.assertEqual(len(distances), 1)
        self.assertEqual(next(iter(distances.values())), 5)


class TestFanout(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_fanout_simple(self):
        i = pyrtl.Input(1, "i")
        o = pyrtl.Output(3, "o")
        w = i ^ 1
        y = i | 1
        z = i & 0
        o <<= w + y + z
        self.assertEqual(pyrtl.fanout(i), 3)
        for wire in (w, y, z):
            self.assertEqual(pyrtl.fanout(wire), 1)
        self.assertEqual(pyrtl.fanout(o), 0)

    def test_fanout_wire_repeated_as_arg(self):
        i = pyrtl.Input(1, "i")
        _w = i * i
        self.assertEqual(pyrtl.fanout(i), 2)

    def test_fanout_wire_repeated_in_concat(self):
        w = pyrtl.WireVector(2)
        _c = pyrtl.concat(w, w, w, w)
        self.assertEqual(pyrtl.fanout(w), 4)


if __name__ == "__main__":
    unittest.main()



import io
import unittest

import pyrtl


class TestConditional(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_basic_true_condition(self):
        c = pyrtl.Const(1)
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with c:
                r.next |= r + 1
        self.check_trace("r 01230123\n")

    def test_short_true_condition(self):
        c = pyrtl.Const(1)
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with c:
                r.next |= r + 1
        self.check_trace("r 01230123\n")

    def test_basic_false_condition(self):
        c = pyrtl.Const(0)
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with c:
                r.next |= r + 1
        self.check_trace("r 00000000\n")

    def test_basic_simple_condition_1(self):
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with r < 2:
                r.next |= r + 1
        self.check_trace("r 01222222\n")

    def test_two_seperate_conditions(self):
        c = pyrtl.Const(1)
        i = pyrtl.Register(bitwidth=2, name="i")
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with c:
                i.next |= i + 1
        with pyrtl.conditional_assignment:
            with i == 2:
                r.next |= r + 1
        self.check_trace("i 01230123\nr 00011112\n")

    def test_basic_two_conditions(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with i == 2:
                r.next |= r + 1
            with i == 3:
                r.next |= r - 1
        self.check_trace("i 01230123\nr 00010001\n")

    def test_basic_default_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r = pyrtl.Register(bitwidth=2, name="r")
        with pyrtl.conditional_assignment:
            with i == 2:
                r.next |= r
            with i == 3:
                r.next |= r - 1
            with pyrtl.otherwise:
                r.next |= r + 1
        self.check_trace("i 01230123\nr 01221233\n")

    def test_basic_nested_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r = pyrtl.Register(bitwidth=3, name="r")
        with pyrtl.conditional_assignment:
            with (i == 2) | (i == 3):
                with r < 3:
                    r.next |= r + 2
                with pyrtl.otherwise:
                    r.next |= r + 1
        self.check_trace("i 01230123\nr 00024445\n")

    def test_nested_under_default_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r = pyrtl.Register(bitwidth=3, name="r")
        with pyrtl.conditional_assignment:
            with i < 2:
                r.next |= r + 2
            with pyrtl.otherwise:
                with r < 6:
                    r.next |= r - 1
        self.check_trace("i 01230123\nr 02432466\n")

    def test_two_signals_under_default_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        with pyrtl.conditional_assignment:
            with i < 2:
                r1.next |= r1 + 1
            with i < 3:
                r2.next |= r2 + 1
            with pyrtl.otherwise:
                r2.next |= 3
        self.check_trace(" i 01230123\nr1 01222344\nr2 00013334\n")

    def test_default_value_for_wires(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        r3 = pyrtl.Register(bitwidth=3, name="r3")
        o = pyrtl.Output(bitwidth=3, name="o")
        with pyrtl.conditional_assignment(defaults={r1: r1 + 2, r2: 6, o: 3}):
            with i < 2:
                r1.next |= r1 + 1
                # r2 will be updated to 6
                # r3 remains the same as previous cycle
                o |= i
            with i < 3:
                # r1 will be updated to r1 + 2
                r2.next |= r2 + 1
                # r3 remains the same as previous cycle
                # o will be updated to 3
            with pyrtl.otherwise:
                # r1 will be updated to r1 + 2
                # r2 will be updated to 6
                r3.next |= 2
                o |= 7
        self.check_trace(
            " i 01230123\n o 01370137\nr1 01246702\nr2 06676667\nr3 00002222\n"
        )

    def test_error_on_unconditioned_update_in_under_conditional(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        with pyrtl.conditional_assignment:
            with self.assertRaises(pyrtl.PyrtlError):
                i.next |= i + 1

    def test_error_on_conditional_assignment_not_under_conditional(self):
        c = pyrtl.Const(1)
        o = pyrtl.Register(bitwidth=2, name="i")
        with self.assertRaises(pyrtl.PyrtlError):
            o |= c

    def test_error_on_non_boolean(self):
        c = pyrtl.Const(2)
        with self.assertRaises(pyrtl.PyrtlError):
            with pyrtl.conditional_assignment:
                with c:
                    pass

    def test_error_on_non_boolean_with_register(self):
        r = pyrtl.Register(bitwidth=4, name="r")
        with self.assertRaises(pyrtl.PyrtlError):
            with pyrtl.conditional_assignment:
                with r:
                    pass


class TestMemConditionalBlock(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_basic_true_condition_memwrite(self):
        m = pyrtl.MemBlock(addrwidth=2, bitwidth=2, name="m")
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i <= 2:
                m[i] |= i
        o <<= m[i]
        self.check_trace("i 01230123\no 00000120\n")

    def test_basic_true_condition_memwrite_2(self):
        m = pyrtl.MemBlock(addrwidth=2, bitwidth=2, name="m")
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with m[i] != 0:
                m[i] <<= i
        o <<= m[i]
        self.check_trace("i 01230123\no 00000123\n")

    def test_basic_true_condition_memread(self):
        m = pyrtl.MemBlock(addrwidth=2, bitwidth=3, name="m")
        i = pyrtl.Register(bitwidth=3, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        addr = i[0:2]
        with pyrtl.conditional_assignment:
            with i < 4:
                m[addr] |= i
            with pyrtl.otherwise:
                with addr[0]:
                    # this should happen every time because no state is being updated!
                    o <<= m[addr]
        self.check_trace("i 01234567\no 00000123\n")

    def test_true_multi_condition_memread(self):
        m = pyrtl.MemBlock(addrwidth=2, bitwidth=3, name="m")
        i = pyrtl.Register(bitwidth=3, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        addr = i[0:2]
        with pyrtl.conditional_assignment:
            with i < 2:
                m[addr] |= i
            with (i >= 2) & (i < 4):
                m[addr] |= (i + 1)[:3]
            with pyrtl.otherwise:
                with addr[0]:
                    # this should happen every time because no state is being updated!
                    o <<= m[addr]
        self.check_trace("i 01234567\no 00000130\n")


class TestWireConditionalBlock(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_basic_condition_wire(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i <= 2:
                o |= 1
            with pyrtl.otherwise:
                o |= 0
        self.check_trace("i 01230123\no 11101110\n")

    def test_boolean_assignment_condition_wire(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i[0]:
                o |= 1
            with pyrtl.otherwise:
                o |= 0
        self.check_trace("i 01230123\no 01010101\n")

    def test_nested_condition_wire(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i <= 2:
                with i == 0:
                    o |= 2
                with pyrtl.otherwise:
                    o |= 1
            with pyrtl.otherwise:
                o |= 0
        self.check_trace("i 01230123\no 21102110\n")

    def test_underspecified_condition_wire(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i <= 2:
                o |= 1
            # default to zero
        self.check_trace("i 01230123\no 11101110\n")

    def test_condition_wire(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        with pyrtl.conditional_assignment:
            with i <= 2:
                o |= 1
        self.check_trace("i 01230123\no 11101110\n")

    def test_condition_nice_error_message_nested(self):
        with pyrtl.conditional_assignment:
            with self.assertRaises(pyrtl.PyrtlError):
                with pyrtl.conditional_assignment:
                    pass

    def test_condition_nice_error_message_nested2(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        with pyrtl.conditional_assignment:
            with i <= 2:
                with self.assertRaises(pyrtl.PyrtlError):
                    with pyrtl.conditional_assignment:
                        pass

    def test_condition_nice_error_message(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        with self.assertRaises(pyrtl.PyrtlError):
            with i <= 2:
                pass

    def test_condition_error_when_assigned_wire_has_unspecified_bitwidth(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(name="o")
        with pyrtl.conditional_assignment:
            with i <= 2:
                with self.assertRaises(pyrtl.PyrtlError):
                    o |= 1

    def test_error_condition_connect_input(self):
        i = pyrtl.Input(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        with pyrtl.conditional_assignment:
            with i <= 2:
                with self.assertRaises(pyrtl.PyrtlError):
                    i |= o

    def test_error_condition_connect_const(self):
        i = pyrtl.Const(3, 2)
        o = pyrtl.WireVector(bitwidth=2, name="o")
        with pyrtl.conditional_assignment:
            with i <= 2:
                with self.assertRaises(pyrtl.PyrtlError):
                    i |= o


class TestNonExclusiveBlocks(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def tearDown(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_basic_nested_non_exclusive_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        with pyrtl.conditional_assignment:
            with r1 < 3:
                r1.next |= r1 + 1
            with pyrtl.otherwise:
                pass
            with r2 < 3:
                r2.next |= r2 + 1
        self.check_trace(" i 01230123\nr1 01233333\nr2 01233333\n")

    def test_one_deep_nested_non_exclusive_condition(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        with pyrtl.conditional_assignment:
            with (i == 2) | (i == 3):
                with r1 < 3:
                    r1.next |= r1 + 2
                with pyrtl.otherwise:
                    pass
                with r2 < 3:
                    r2.next |= r2 + 2
        self.check_trace(" i 01230123\nr1 00024444\nr2 00024444\n")

    def test_overlaping_assignments_in_non_exclusive_assignments(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        with pyrtl.conditional_assignment:
            with r1 < 3:
                r1.next |= r1 + 1
            with pyrtl.otherwise:
                pass
            with r2 < 3:
                with self.assertRaises(pyrtl.PyrtlError):
                    r1.next |= r2 + 1

    def test_other_overlaping_assignments_in_non_exclusive_assignments(self):
        i = pyrtl.Register(bitwidth=2, name="i")
        i.next <<= i + 1
        r1 = pyrtl.Register(bitwidth=3, name="r1")
        r2 = pyrtl.Register(bitwidth=3, name="r2")
        with pyrtl.conditional_assignment:
            with r1 < 3:
                r1.next |= r1 + 1
            with pyrtl.otherwise:
                pass
            with r2 < 3:
                with i != 0:
                    with self.assertRaises(pyrtl.PyrtlError):
                        r1.next |= r2 + 1


class TestSuperWireConditionalBlock(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_super_stress_test(self):
        allin = [pyrtl.Input(1, n) for n in "abxyijkmzcd"]
        a, b, x, y, i, j, k, m, z, c, d = allin
        allout = [pyrtl.Output(4, "var" + str(index)) for index in range(5)]
        var0, var1, var2, var3, var4 = allout

        """
         1  with a:  # a
         2  with b:  # not(a) and b
         3      with x:  # not(a) and b and x
         4      with otherwise:  # not(a) and b and not(x)
         5      with y:  # not(a) and b and y;  check(3,4)
         6          with i:  # not(a) and b and y and i;  check(3,4)
         7          with j:  # not(a) and b and y and not(i) and j;  check(3,4)
         8          with otherwise:  # not(a) and b and y and not(i) and not(j):
                                                                     check(3,4)
         9          with k:  # not(a) and b and y and k;  check(3,4,6,7,8)
         8          with otherwise:  # not(a) and b and y and not(k):  check(?)
        10          with m:  # not(a) and b and y and m;  check(?)
        11  with otherwise:  #not(a) and not(b)
        12      with z: #not(a) and not(b) and z
        13  with c:  #c;  check(1,2,3,4,5,6,7,8,9,10,11,12)
        14  with d:  #not(c) and d;  check(1,2,3,4,5,6,7,8,9,10,11,12)
        """

        with pyrtl.conditional_assignment:
            with a:
                var0 |= 1
            with b:
                with x:
                    var0 |= 2
                with pyrtl.otherwise:
                    var0 |= 3

                with y:
                    with i:
                        var1 |= 1
                    with j:
                        var1 |= 2
                    with pyrtl.otherwise:
                        var1 |= 3

                    with k:
                        var2 |= 1
                    with pyrtl.otherwise:
                        var2 |= 2

                    with m:
                        var3 |= 3

            with pyrtl.otherwise:
                with z:
                    var0 |= 4

            with c:
                var4 |= 1
            with d:
                var4 |= 2

        sim = pyrtl.FastSimulation()
        for cycle in range(2 ** len(allin)):
            inputs = {v: 0x1 & (cycle >> i) for i, v in enumerate(allin[::-1])}
            sim.step(inputs)

        for cycle in range(len(sim.tracer)):
            t0, t1, t2, t3, t4 = 0, 0, 0, 0, 0

            def v(var, cycle=cycle):
                return sim.tracer.trace[var.name][cycle]

            if v(a):
                t0 = 1
            elif v(b):
                if v(x):
                    t0 = 2
                else:
                    t0 = 3
                if v(y):
                    if v(i):
                        t1 = 1
                    elif v(j):
                        t1 = 2
                    else:
                        t1 = 3
                    if v(k):
                        t2 = 1
                    else:
                        t2 = 2
                    if v(m):
                        t3 = 3
            else:
                if v(z):
                    t0 = 4
            if v(c):
                t4 = 1
            elif v(d):
                t4 = 2
            self.assertEqual(v(var0), t0)
            self.assertEqual(v(var1), t1)
            self.assertEqual(v(var2), t2)
            self.assertEqual(v(var3), t3)
            self.assertEqual(v(var4), t4)


if __name__ == "__main__":
    unittest.main()



import random
import unittest
from itertools import islice

import pyrtl
from pyrtl.rtllib import prngs


class TestPrngs(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    @staticmethod
    def fibonacci_lfsr(seed):
        lfsr = seed
        while True:
            bit = (lfsr >> 126 ^ lfsr >> 125) & 1
            lfsr = lfsr << 1 | bit
            yield bit

    @staticmethod
    def xoroshiro128plus(seed):
        mask = 2**64 - 1
        s = [seed & mask, seed >> 64]
        while True:
            s0 = s[0]
            s1 = s[1]
            word = (s1 + s0) & mask
            s1 ^= s0
            s[0] = (s0 << 55 | s0 >> 9) ^ s1 ^ s1 << 14
            s[1] = s1 << 36 | s1 >> 28
            yield word

    def test_prng_lfsr(self):
        seed = pyrtl.Input(127, "seed")
        load, req = pyrtl.Input(1, "load"), pyrtl.Input(1, "req")
        rand = pyrtl.Output(128, "rand")
        rand <<= prngs.prng_lfsr(128, load, req, seed)
        sim = pyrtl.Simulation()
        in_vals = [random.randrange(1, 2**127) for i in range(5)]

        for trial in range(5):
            true_val = 0
            for bit in islice(TestPrngs.fibonacci_lfsr(in_vals[trial]), 128):
                true_val = true_val << 1 | bit
            sim.step({"load": 1, "req": 0, "seed": in_vals[trial]})
            sim.step({"load": 0, "req": 1, "seed": 0x0})
            sim.step({"load": 0, "req": 0, "seed": 0x0})
            circuit_out = sim.inspect(rand)
            self.assertEqual(
                circuit_out,
                true_val,
                f"\nAssertion failed on trial {trial}\n"
                f"Expected value: {hex(true_val)}\nGotten value: {hex(circuit_out)}",
            )

    def test_prng_xoroshiro128(self):
        seed = pyrtl.Input(128, "seed")
        load, req = pyrtl.Input(1, "load"), pyrtl.Input(1, "req")
        ready = pyrtl.Output(1, "ready")
        rand = pyrtl.Output(128, "rand")
        ready_out, rand_out = prngs.prng_xoroshiro128(128, load, req, seed)
        ready <<= ready_out
        rand <<= rand_out
        sim = pyrtl.Simulation()
        in_vals = [random.randrange(1, 2**128) for i in range(5)]

        for trial in range(5):
            true_val = 0
            for word in islice(TestPrngs.xoroshiro128plus(in_vals[trial]), 2):
                true_val = true_val << 64 | word
            sim.step({"load": 1, "req": 0, "seed": in_vals[trial]})
            sim.step({"load": 0, "req": 1, "seed": 0x0})
            for _cycle in range(2, 4):
                sim.step({"load": 0, "req": 0, "seed": 0x0})
            circuit_out = sim.inspect(rand)
            self.assertEqual(
                circuit_out,
                true_val,
                f"\nAssertion failed on trial {trial}\n"
                f"Expected value: {hex(true_val)}\nGotten value: {hex(circuit_out)}",
            )

        for ready_signal in sim.tracer.trace["ready"][:3]:
            self.assertEqual(ready_signal, 0)
        self.assertEqual(sim.tracer.trace["ready"][3], 1)
        self.assertEqual(sim.tracer.trace["ready"][4], 0)

    def test_csprng_trivium(self):
        """
        Trivium test vectors retrived from:
        https://www.sheffield.ac.uk/polopoly_fs/1.12164!/file/eSCARGOt_full_datasheet_v1.3.pdf
        bit ordering is modified to adapt to the Pyrtl implementation
        """
        in_vector = pyrtl.Input(160, "in_vector")
        load, req = pyrtl.Input(1, "load"), pyrtl.Input(1, "req")
        ready = pyrtl.Output(1, "ready")
        out_vector = pyrtl.Output(128, "out_vector")
        ready_out, rand_out = prngs.csprng_trivium(128, load, req, in_vector)
        ready <<= ready_out
        out_vector <<= rand_out
        sim = pyrtl.Simulation()

        in_vals = [
            0x0100000000000000000000000000000000000000,
            0x0A09080706050403020100000000000000000000,
            0xFFFEFDFCFBFAF9F8F7F600000000000000000000,
            0xFAA75401AE5B08B5620FC760F9922BC45DF68F28,
            0xF5A24FFCA95603B05D0ABE57F08922BB54ED861F,
        ]

        true_vals = [
            0x1CD761FFCEB05E39F5B18F5C22042AB0,
            0x372E6B86524AFA71B5FEE86D5CEBB07D,
            0xC100BACA274287277FF49B9FB512AF1C,
            0xCB5996FCFF373A953FC169E899E02F46,
            0xF142D1DF4B36C7652CBA2E4A22EE51A0,
        ]

        for trial in range(5):
            sim.step({"load": 1, "req": 0, "in_vector": in_vals[trial]})
            for _cycle in range(1, 20):
                sim.step({"load": 0, "req": 0, "in_vector": 0x0})
            sim.step({"load": 0, "req": 1, "in_vector": 0x0})
            for _cycle in range(21, 23):
                sim.step({"load": 0, "req": 0, "in_vector": 0x0})
            circuit_out = sim.inspect(out_vector)
            self.assertEqual(
                circuit_out,
                true_vals[trial],
                f"\nAssertion failed on trial {trial}\n"
                f"Expected value: {hex(true_vals[trial])}\n"
                f"Gotten value: {hex(circuit_out)}",
            )

        for ready_signal in sim.tracer.trace["ready"][:19]:
            self.assertEqual(ready_signal, 0)
        self.assertEqual(sim.tracer.trace["ready"][19], 1)

        for ready_signal in sim.tracer.trace["ready"][20:22]:
            self.assertEqual(ready_signal, 0)
        self.assertEqual(sim.tracer.trace["ready"][22], 1)
        self.assertEqual(sim.tracer.trace["ready"][23], 0)


if __name__ == "__main__":
    unittest.main()



# Most users should only interact with the objects "conditional_assignment" and
# "otherwise". The classes defined below are internal implementation details.

from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.wire import Const, Register, WireVector

# -----------------------------------------------------------------------
#    __   __        __    ___    __                  __
#   /  ` /  \ |\ | |  \ |  |  | /  \ |\ |  /\  |    /__`
#   \__, \__/ | \| |__/ |  |  | \__/ | \| /~~\ |___ .__/
#


def currently_under_condition() -> bool:
    """
    :return: ``True`` iff execution is currently in the context of a
        :data:`conditional_assignment`.

    """
    return _depth > 0


# `conditional_assignment` and `otherwise`, both visible in the pyrtl module, are
# defined as instances (hopefully the only and unchanging instances) of the following
# two types.
class _ConditionalAssignment:
    def __init__(self):
        self.defaults = {}

    def __call__(self, defaults):
        self.defaults = defaults
        return self

    """ Context providing functionality of "conditional_assignment". """

    def __enter__(self):
        global _depth
        _check_no_nesting()
        _depth = 1

    def __exit__(self, *exc_info):
        try:
            _finalize(self.defaults)
        finally:
            # even if the above finalization throws an error we need to reset the state
            # to prevent errors from bleeding over
            _reset_conditional_state()  # sets _depth back to 0


class _Otherwise:
    def __enter__(self):
        _push_condition(otherwise)

    def __exit__(self, *exc_info):
        _pop_condition()


def _reset_conditional_state():
    """Set or reset all the module state required for conditionals."""
    global _conditions_list_stack
    global _conflicts_map
    global _predicate_map
    global _depth
    _depth = 0
    _conditions_list_stack = [[]]  # stack of lists of current conditions
    # _predicate_map: map wirevector or mem -> [(final_pred, rhs), ...]
    _predicate_map = {}
    # _conflicts_map: map wirevector or mem ->
    #     [ set([(pred,bool), (pred,bool)]), set([(pred,bool)..
    # * each element maps to a list of sets of tuples of (predicate id, bool)
    # * each time a value is written (lhs) we add the predicate set to the list
    # * each new write happens we have to check that the new predicate has at least one
    #   negated term with the value we are now trying to write. Otherwise it is an
    #   error.
    _conflicts_map = {}


_reset_conditional_state()
conditional_assignment = _ConditionalAssignment()

otherwise = _Otherwise()


def _push_condition(predicate):
    """As we enter new conditions, this pushes them on the predicate stack."""
    global _depth
    _check_under_condition()
    _depth += 1
    if predicate is not otherwise and len(predicate) > 1:
        msg = "all predicates for conditional assignments must be wirevectors of len 1"
        raise PyrtlError(msg)
    _conditions_list_stack[-1].append(predicate)
    _conditions_list_stack.append([])


def _pop_condition():
    """As we exit conditions, this pops them off the stack."""
    global _depth
    _check_under_condition()
    _conditions_list_stack.pop()
    _depth -= 1


def _build(lhs, rhs):
    """Stores the wire assignment details until finalize is called."""
    _check_under_condition()
    final_predicate, pred_set = _current_select()
    _check_and_add_pred_set(lhs, pred_set)
    _predicate_map.setdefault(lhs, []).append((final_predicate, rhs))


def _build_read_port(mem, addr):
    # TODO: reduce number of ports through collapsing reads
    return mem._build_read_port(addr)


def _check_no_nesting():
    if _depth != 0:
        msg = "no nesting of conditional assignments allowed"
        raise PyrtlError(msg)


def _check_under_condition():
    if not currently_under_condition():
        msg = 'conditional assignment "|=" only valid under a condition'
        raise PyrtlError(msg)


def _check_and_add_pred_set(lhs, pred_set):
    for test_set in _conflicts_map.setdefault(lhs, []):
        if _pred_sets_are_in_conflict(pred_set, test_set):
            msg = f"conflicting conditions for {lhs}"
            raise PyrtlError(msg)
    _conflicts_map[lhs].append(pred_set)


def _pred_sets_are_in_conflict(pred_set_a, pred_set_b):
    """Find conflict in sets, return conflict if found, else None."""
    # pred_sets conflict if we cannot find one shared predicate that is "negated" in one
    # and "non-negated" in the other
    for pred_a, bool_a in pred_set_a:
        for pred_b, bool_b in pred_set_b:
            if pred_a is pred_b and bool_a != bool_b:
                return False
    return True


def _finalize(defaults):
    """
    Build the required muxes and call back to WireVector to finalize the wirevector
    build.
    """
    from pyrtl.corecircuits import select
    from pyrtl.memory import MemBlock

    for lhs in _predicate_map:
        # handle memory write ports
        if isinstance(lhs, MemBlock):
            p, (addr, data, enable) = _predicate_map[lhs][0]
            combined_enable = select(p, truecase=enable, falsecase=Const(0))
            combined_addr = addr
            combined_data = data

            for p, (addr, data, enable) in _predicate_map[lhs][1:]:
                combined_enable = select(p, truecase=enable, falsecase=combined_enable)
                combined_addr = select(p, truecase=addr, falsecase=combined_addr)
                combined_data = select(p, truecase=data, falsecase=combined_data)

            lhs._build(combined_addr, combined_data, combined_enable)

        # handle wirevector and register assignments
        else:
            if isinstance(lhs, Register):
                if lhs in defaults:
                    result = defaults[lhs]
                else:
                    result = lhs  # default for registers is "self"
            elif isinstance(lhs, WireVector):
                if lhs in defaults:
                    result = defaults[lhs]
                else:
                    result = 0  # default for wire is "0"
            else:
                msg = "unknown assignment in finalize"
                raise PyrtlInternalError(msg)
            predlist = _predicate_map[lhs]
            for p, rhs in predlist:
                result = select(p, truecase=rhs, falsecase=result)
            lhs._build(result)


def _current_select():
    """Function to calculate the current "predicate" in the current context.

    Returns a tuple of information: (predicate, pred_set). The value pred_set is a set([
    (predicate, bool), ... ]) as described in the _reset_conditional_state
    """

    # helper to create the conjuction of predicates
    def and_with_possible_none(a, b):
        assert a is not None or b is not None
        if a is None:
            return b
        if b is None:
            return a
        return a & b

    def between_otherwise_and_current(predlist):
        lastother = None
        for i, p in enumerate(predlist[:-1]):
            if p is otherwise:
                lastother = i
        if lastother is None:
            return predlist[:-1]
        return predlist[lastother + 1 : -1]

    select = None
    pred_set = set()

    # for all conditions except the current children (which should be [])
    for predlist in _conditions_list_stack[:-1]:
        # negate all of the predicates between "otherwise" and the current one
        for predicate in between_otherwise_and_current(predlist):
            select = and_with_possible_none(select, ~predicate)
            pred_set.add((predicate, True))
        # include the predicate for the current one (not negated)
        if predlist[-1] is not otherwise:
            predicate = predlist[-1]
            select = and_with_possible_none(select, predicate)
            pred_set.add((predicate, False))

    if select is None:
        msg = "problem with conditional assignment"
        raise PyrtlError(msg)
    if len(select) != 1:
        msg = "conditional predicate with length greater than 1"
        raise PyrtlInternalError(msg)

    return select, pred_set


# Some examples that were helpful in the design and testing of conditional

#  1  with a:  # a
#  2  with b:  # not(a) and b
#  3    with x:  # not(a) and b and x
#  4    with otherwise:  # not(a) and b and not(x)
#  5    with y:  # not(a) and b and y;  check(3,4)
#  6        with i:  # not(a) and b and y and i;  check(3,4)
#  7        with j:  # not(a) and b and y and not(i) and j;  check(3,4)
#  8        with otherwise:  # not(a) and b and y and not(i) and not(j):  check(3,4)
#  9        with k:  # not(a) and b and y and k;  check(3,4,6,7,8)
# 10        with m:  # not(a) and b and y and not(k) and m;  check(3,4,6,7,8)
# 11  with otherwise:  #not(a) and not(b)
# 12  with c:  #c;  check(1,2,3,4,5,6,7,8,9,10,11)

#  0  with a:  # a
#  1  with otherwise:  # a;
#  2  with b:  # not(a) and b;  check(0,1)
#  3    with x:  # not(a) and b and x;  check(0,1)
#  4    with otherwise:  # not(a) and b and not(x);  check(0,1)
#  5    with y:  # not(a) and b and y;  check(0,1,3,4)
#  6        with i:  # not(a) and b and y and i;  check(0,1,3,4)
#  7        with j:  # not(a) and b and y and not(i) and j;  check(0,1,3,4)
#  8        with otherwise:  # not(a) and b and y and not(i) and not(j):  check(0,1,3,4)
#  9        with k:  # not(a) and b and y and k;  check(0,1,3,4,6,7,8)
# 10        with m:  # not(a) and b and y and not(k) and m;  check(0,1,3,4,6,7,8)
#       with z: check(0,1,3,4)
#       with otherwise: check(0,1,3,4)
#       with g: check(0,1,3,4,5,6,7,8,9,10)
# 11  with otherwise:  #not(a) and not(b);  check(0,1)
# 12  with c:  #c;  check(0,1,2,3,4,5,6,7,8,9,10,11)



# # Example 8: Interfacing with Verilog.
#
# PyRTL provides utilities to import from, and export to, Verilog. Simulation traces can
# also be exported to `vcd` files.
import io
import random

import pyrtl

# ## Importing From Verilog
#
# Sometimes it is useful to pull in components written in Verilog to be used as
# subcomponents of PyRTL designs, or for analysis in PyRTL. PyRTL supports the standard
# `blif` format: https://www.ece.cmu.edu/~ee760/760docs/blif.pdf
#
# Many tools support outputting hardware designs to `blif`, including the open source
# project `yosys`. `blif` files can then be imported either as a string or directly from
# a file name by the function `input_from_blif()`. Here is a simple example of a 1-bit
# full adder imported and then simulated from `blif`:
full_adder_blif = """
.model full_adder
.inputs x y cin
.outputs sum cout
.names $false
.names $true
1
.names y $not$FA.v:12$3_Y
0 1
.names x $not$FA.v:11$1_Y
0 1
.names cin $not$FA.v:15$6_Y
0 1
.names ind3 ind4 sum
1- 1
-1 1
.names $not$FA.v:15$6_Y ind2 ind3
11 1
.names x $not$FA.v:12$3_Y ind1
11 1
.names ind2 $not$FA.v:16$8_Y
0 1
.names cin $not$FA.v:16$8_Y ind4
11 1
.names x y $and$FA.v:19$11_Y
11 1
.names ind0 ind1 ind2
1- 1
-1 1
.names cin ind2 $and$FA.v:19$12_Y
11 1
.names $and$FA.v:19$11_Y $and$FA.v:19$12_Y cout
1- 1
-1 1
.names $not$FA.v:11$1_Y y ind0
11 1
.end
"""

pyrtl.input_from_blif(full_adder_blif)
# Find the `WireVectors` corresponding to wires named in the `blif` file.
x, y, cin = [pyrtl.working_block().get_wirevector_by_name(s) for s in ["x", "y", "cin"]]

# Simulate the logic with random input values:
sim = pyrtl.Simulation()
for _cycle in range(15):
    sim.step(
        {"x": random.randrange(2), "y": random.randrange(2), "cin": random.randrange(2)}
    )
# Only display the `Input` and `Output` `WireVectors` for clarity.
input_vectors = pyrtl.working_block().wirevector_subset(pyrtl.Input)
output_vectors = pyrtl.working_block().wirevector_subset(pyrtl.Output)
sim.tracer.render_trace(trace_list=[*input_vectors, *output_vectors], symbol_len=2)

# ## Exporting to Verilog
#
# To demonstrate Verilog export, we create a simple 3-bit counter. This is like the
# counter in `example2`, except this one can be reset at any time by asserting `zero`.
pyrtl.reset_working_block()

zero = pyrtl.Input(1, "zero")
counter_output = pyrtl.Output(3, "counter_output")
counter = pyrtl.Register(3, "counter")
counter.next <<= pyrtl.select(zero, 0, counter + 1)
counter_output <<= counter

# The `select()` statement resets the `counter` to `0` in the next cycle when the `zero`
# signal goes high, otherwise the `counter`'s next value is just `counter + 1`.
#
# The constants `0` and `1` are automatically `zero_extended()` to the proper length.
# Let's export this resettable counter to a Verilog file and see what is looks like
# (here we are using `StringIO` just to print it to a string for demo purposes; most
# likely you will want to pass a normal `open` file).
print("\n--- PyRTL Representation ---")
print(pyrtl.working_block())

print("\n--- Verilog for the Counter ---")
with io.StringIO() as verilog_file:
    pyrtl.output_to_verilog(verilog_file)
    print(verilog_file.getvalue())

print("\n--- Simulation Results ---")
sim = pyrtl.Simulation(tracer=pyrtl.SimulationTrace([counter_output, zero]))
for _cycle in range(15):
    sim.step({"zero": random.choice([0, 0, 0, 1])})
sim.tracer.render_trace()

# We already did the "hard" work of generating a test input for this simulation, so we
# might want to reuse that work when we take this design through a Verilog toolchain.
# `output_verilog_testbench()` grabs the `Inputs` used in the `SimulationTrace` and sets
# them up in a standard Verilog testbench.
print("\n--- Verilog for the TestBench (first 10 lines) ---")
with io.StringIO() as testbench_file:
    pyrtl.output_verilog_testbench(
        dest_file=testbench_file, simulation_trace=sim.tracer
    )
    for i, line in enumerate(testbench_file.getvalue().split("\n")):
        if i == 10:
            break
        print(line)
    print("...")

# ## Transformations
#
# Now let's talk about transformations of the hardware block. Many times when you are
# doing some hardware-level analysis you might wish to ignore higher level things like
# multi-bit wirevectors, adds, concatenation, etc, and just think about wires and basic
# gates. PyRTL supports "lowering" of designs into this more restricted set of
# functionality though the function `synthesize()`. Once we lower a design to this form
# we can then apply basic optimizations like constant propagation and dead wire
# elimination as well. By printing it out to Verilog we can see exactly how the design
# changed.
print("\n--- Optimized Single-bit Verilog for the Counter ---")
pyrtl.synthesize()
pyrtl.optimize()

with io.StringIO() as verilog_file:
    pyrtl.output_to_verilog(verilog_file)
    print(verilog_file.getvalue())
